import logging
import syslog
# syslog.syslog("### gulliversoft: This is a test message")
# syslog.syslog(syslog.LOG_INFO, "### gulliversoft: Test message at INFO priority")
# syslog.syslog(syslog.LOG_DEBUG, "### gulliversoft: Test message at DEBUG priority")

LOG_FORMATTER = logging.Formatter(fmt='%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt="%Y-%m-%d %H:%M:%S")        
ROOT_LOGGER = logging.getLogger()
ROOT_LOGGER.setLevel(logging.WARN)

CONSOLE_HANDLER = logging.StreamHandler()
CONSOLE_HANDLER.setFormatter(LOG_FORMATTER)
ROOT_LOGGER.addHandler(CONSOLE_HANDLER)

LOGGER = logging.getLogger("Daisy")

def setInfo():
    ROOT_LOGGER.setLevel(logging.INFO)

def setDebug():
    ROOT_LOGGER.setLevel(logging.DEBUG)
    
def debugEnabled():
    return ROOT_LOGGER.level == logging.DEBUG

def logToFile(filename):
    FILE_HANDLER = logging.FileHandler(filename)
    FILE_HANDLER.setFormatter(LOG_FORMATTER)
    ROOT_LOGGER.addHandler(FILE_HANDLER)

def debug(message):
    LOGGER.debug(message)
    syslog.syslog(syslog.LOG_DEBUG, message)

def info(message):
    LOGGER.info(message)
    syslog.syslog(syslog.LOG_INFO, message)

def warn(message):
    LOGGER.warn(message)
    syslog.syslog(syslog.LOG_WARNING, message)

def error(message):
    LOGGER.error(message)
    syslog.syslog(syslog.LOG_ERR, message)

def exception(message):
    LOGGER.exception(message)
    syslog.syslog(syslog.LOG_ALERT, message)

def printBytes(buff):
    for i in range(0, len(buff)):
        print("%03d: 0x%02X %03d %c" % (i, buff[i], buff[i], buff[i]))
        




from rocket.daisy.utils.types import toint, signInteger
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.sensor import Temperature

class TMP102(I2C, Temperature):
    def __init__(self, slave=0x48):
        I2C.__init__(self, toint(slave))
        
    def __str__(self):
        return "TMP102(slave=0x%02X)" % self.slave
        
    def __getKelvin__(self):
        return self.Celsius2Kelvin()

    def __getCelsius__(self):
        d = self.readBytes(2)
        count = ((d[0] << 4) | (d[1] >> 4)) & 0xFFF
        return signInteger(count, 12)*0.0625
    
    def __getFahrenheit__(self):
        return self.Celsius2Fahrenheit()

class TMP75(TMP102):
    def __init__(self, slave=0x48, resolution=12):
        TMP102.__init__(self, slave)
        resolution = toint(resolution)
        if not resolution in range(9,13):
            raise ValueError("%dbits resolution out of range [%d..%d]bits" % (resolution, 9, 12))
        self.resolution = resolution
        
        config  = self.readRegister(0x01)
        config &= ~0x60
        config |= (self.resolution - 9) << 5
        self.writeRegister(0x01, config)
        self.readRegisters(0x00, 2)

    def __str__(self):
        return "TMP75(slave=0x%02X, resolution=%d-bits)" % (self.slave, self.resolution)
        
class TMP275(TMP75):
    def __init__(self, slave=0x48, resolution=12):
        TMP75.__init__(self, slave, resolution)

    def __str__(self):
        return "TMP275(slave=0x%02X, resolution=%d-bits)" % (self.slave, self.resolution)
        




from rocket.daisy.utils.types import toint
from rocket.daisy.devices.spi import SPI
from rocket.daisy.devices.analog import DAC

class MCP48XX(SPI, DAC):
    def __init__(self, chip, channelCount, resolution, name):
        SPI.__init__(self, toint(chip), 0, 8, 10000000)
        DAC.__init__(self, channelCount, resolution, 2.048)
        self.name = name
        self.buffered=False
        self.gain=False
        self.shutdown=True
        self.values = [0 for i in range(channelCount)]

    def __str__(self):
        return "%s(chip=%d)" % (self.name, self.chip)

    def int2bin(self,n,count):
        return "".join([str((n >> y) & 1 ) for y in range(count-1,-1,-1)])

    def __analogRead__(self, channel, diff=False):
        return self.values[channel]

    def __analogWriteShut__(self, channel):
        self.shutdown = True
        d = [0x00, 0x00]
        d[0] |= (channel & 0x01) << 7             # bit 15 = channel
        d[0] |= 0x00 << 6                         # bit 14 = ignored
        d[0] |= 0x00 << 5                         # bit 13 = gain
        d[0] |= (self.shutdown & 0x01) << 4       # bit 12 = shutdown
        d[0] |= 0x00                              # bits 8-11 = msb data
        d[1] |= 0x00                              # bits 0 - 7 = lsb data
        
        self.writeBytes(d)
        self.values[channel] = 0

    def __analogWrite__(self, channel, value):
        self.shutdown=False
        d = [0x00, 0x00]
        d[0] |= (channel & 0x01) << 7                     # bit 15 = channel
        d[0] |= (self.buffered & 0x01) << 6               # bit 14 = ignored
        d[0] |= (not self.gain & 0x01) << 5               # bit 13 = gain
        d[0] |= (not self.shutdown & 0x01) << 4           # bit 12 = shutdown
        d[0] |= value >> (self._analogResolution - 4)     # bits 8-11 = msb data
        d[1] |= ((value << (12-self._analogResolution)) & 0xFF) # bits 4 - 7 = lsb data (4802) bits 2-7 (4812) bits 0-7 (4822)                              # bits 0 - 3 = ignored                       
        
        self.writeBytes(d)
        self.values[channel] = value
       
class MCP4802(MCP48XX):
    def __init__(self, chip=0):
        MCP48XX.__init__(self, chip, 2, 8, "MCP4802")

class MCP4812(MCP48XX):
    def __init__(self, chip=0):
        MCP48XX.__init__(self, chip, 2, 10, "MCP4812")

class MCP4822(MCP48XX):
    def __init__(self, chip=0):
        MCP48XX.__init__(self, chip, 2, 12, "MCP4822")




#     ____        _           
#    / __ \____ _(_)______  __
#   / / / / __ `/ / ___/ / / /
#  / /_/ / /_/ / (__  ) /_/ / 
# /_____/\__,_/_/____/\__, /  
# martin shishkov    /____/   
#
#   Copyright 2020 Martin Shishkov, gulliversoft.com
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from rocket.daisy.utils.logger import LOGGER
from rocket.daisy.utils.version import PYTHON_MAJOR
from rocket.daisy.utils.crypto import encodeCredentials
from rocket.daisy.protocols.coap import COAPClient, COAPGet, COAPPost, COAPPut, COAPDelete

if PYTHON_MAJOR >= 3:
    import http.client as httplib
else:
    import httplib

class PiMixedClient():
    def __init__(self, host, port=8000, coap=5683):
        self.host = host
        if coap > 0:
            self.coapport = coap
            self.coapclient = COAPClient()
        else:
            self.coapclient = None
        if port > 0:
            self.httpclient = httplib.HTTPConnection(host, port)
        else:
            self.httpclient = None
        self.forceHttp = False
        self.coapfailure = 0
        self.maxfailure = 2
        self.auth= None;
    
    def setCredentials(self, login, password):
        self.auth = "Basic " + encodeCredentials(login, password)
        
    def sendRequest(self, method, uri):
        if self.coapclient != None and not self.forceHttp:
            if method == "GET":
                response = self.coapclient.sendRequest(COAPGet("coap://%s:%d%s" % (self.host, self.coapport, uri)))
            elif method == "POST":
                response = self.coapclient.sendRequest(COAPPost("coap://%s:%d%s" % (self.host, self.coapport, uri)))

            if response:
                return str(response.payload)
            elif self.httpclient != None:
                self.coapfailure += 1
                print("No CoAP response, fall-back to HTTP")
                if (self.coapfailure > self.maxfailure):
                    self.forceHttp = True
                    self.coapfailure = 0
                    print("Too many CoAP failure forcing HTTP")
        
        if self.httpclient != None:
            headers = {}
            if self.auth != None:
                headers["Authorization"] = self.auth
            
            self.httpclient.request(method, uri, None, headers)
            response = self.httpclient.getresponse()
            if response.status == 200:
                data = response.read()
                return data
            elif response.status == 401:
                raise Exception("Missing credentials")
            else:
                raise Exception("Unhandled HTTP Response %d %s" % (response.status, response.reason))

        raise Exception("No data received")

class PiHttpClient(PiMixedClient):
    def __init__(self, host, port=8000):
        PiMixedClient.__init__(self, host, port, -1)

class PiCoapClient(PiMixedClient):
    def __init__(self, host, port=5683):
        PiMixedClient.__init__(self, host, -1, port)

class PiMulticastClient(PiMixedClient):
    def __init__(self, port=5683):
        PiMixedClient.__init__(self, "224.0.1.123", -1, port)

class RESTAPI():
    def __init__(self, client, path):
        self.client = client
        self.path = path
        
    def sendRequest(self, method, path):
        return self.client.sendRequest(method, self.path + path)
        
class Macro(RESTAPI):
    def __init__(self, client, name):
        RESTAPI.__init__(self, client, "/macros/" + name + "/")
        
    def call(self, *args):
        values = ",".join(["%s" % i for i in args])
        if values == None:
            values = ""
        return self.sendRequest("POST", values)

class Device(RESTAPI):
    def __init__(self, client, name, category):
        RESTAPI.__init__(self, client, "/devices/" + name + "/" + category)

class GPIO(Device):
    def __init__(self, client, name):
        Device.__init__(self, client, name, "digital")
        
    def getFunction(self, channel):
        return self.sendRequest("GET", "/%d/function" % channel)

    def setFunction(self, channel, func):
        return self.sendRequest("POST", "/%d/function/%s" % (channel, func))
        
    def digitalRead(self, channel):
        return int(self.sendRequest("GET", "/%d/value" % channel))

    def digitalWrite(self, channel, value):
        return int(self.sendRequest("POST", "/%d/value/%d" % (channel, value)))
    
    def portRead(self):
        return int(self.sendRequest("GET", "/integer"))

    def portWrite(self, value):
        return int(self.sendRequest("POST", "/integer/%d" % value))

class NativeGPIO(GPIO):
    def __init__(self, client):
        RESTAPI.__init__(self, client, "/GPIO")

class ADC(Device):
    def __init__(self, client, name):
        Device.__init__(self, client, name, "analog")
        
    def read(self, channel):
        return float(self.sendRequest("GET", "/%d/integer" % channel))

    def readFloat(self, channel):
        return float(self.sendRequest("GET", "/%d/float" % channel))

    def readVolt(self, channel):
        return float(self.sendRequest("GET", "/%d/volt" % channel))

class DAC(ADC):
    def __init__(self, client, name):
        Device.__init__(self, client, name, "analog")
        
    def write(self, channel, value):
        return float(self.sendRequest("POST", "/%d/integer/%d" % (channel, value)))
                     
    def writeFloat(self, channel, value):
        return float(self.sendRequest("POST", "/%d/float/%f" % (channel, value)))
                     
    def writeVolt(self, channel, value):
        return float(self.sendRequest("POST", "/%d/volt/%f" % (channel, value)))
                     
class PWM(DAC):
    def __init__(self, client, name):
        Device.__init__(self, client, name, "pwm")
        
    def readAngle(self, channel, value):
        return float(self.sendRequest("GET", "/%d/angle" % (channel)))
                     
    def writeAngle(self, channel, value):
        return float(self.sendRequest("POST", "/%d/angle/%f" % (channel, value)))
                     
class Sensor(Device):
    def __init__(self, client, name):
        Device.__init__(self, client, name, "sensor")
        
class Temperature(Sensor):
    def getKelvin(self):
        return float(self.sendRequest("GET", "/temperature/k"))

    def getCelsius(self):
        return float(self.sendRequest("GET", "/temperature/c"))

    def getFahrenheit(self):
        return float(self.sendRequest("GET", "/temperature/f"))
    
class Pressure(Sensor):
    def getPascal(self):
        return float(self.sendRequest("GET", "/pressure/pa"))

    def getHectoPascal(self):
        return float(self.sendRequest("GET", "/pressure/hpa"))
    
class Luminosity(Sensor):
    def getLux(self):
        return float(self.sendRequest("GET", "/luminosity/lux"))
    
class Distance(Sensor):
    def getMillimeter(self):
        return float(self.sendRequest("GET", "/distance/mm"))

    def getCentimeter(self):
        return float(self.sendRequest("GET", "/distance/cm"))

    def getInch(self):
        return float(self.sendRequest("GET", "/distance/in"))






from rocket.daisy.utils.version import PYTHON_MAJOR
from rocket.daisy.utils.logger import info, exception 

import socket
import struct
import logging
import threading

M_PLAIN = "text/plain"
M_JSON  = "application/json"

if PYTHON_MAJOR >= 3:
    from urllib.parse import urlparse
else:
    from urlparse import urlparse

def HTTPCode2CoAPCode(code):
    return int(code/100) * 32 + (code%100)

   
class COAPContentFormat():
    FORMATS = {0: "text/plain",
               40: "application/link-format",
               41: "application/xml",
               42: "application/octet-stream",
               47: "application/exi",
               50: "application/json"
               }

    @staticmethod
    def getCode(fmt):
        if fmt == None:
            return None
        for code in COAPContentFormat.FORMATS:
            if COAPContentFormat.FORMATS[code] == fmt:
                return code
        return None
    
    @staticmethod
    def toString(code):
        if code == None:
            return None

        if code in COAPContentFormat.FORMATS:
            return COAPContentFormat.FORMATS[code]
        
        raise Exception("Unknown content format %d" % code)
        

class COAPOption():
    OPTIONS = {1: "If-Match",
               3: "Uri-Host",
               4: "ETag",
               5: "If-None-Match",
               7: "Uri-Port",
               8: "Location-Path",
               11: "Uri-Path",
               12: "Content-Format",
               14: "Max-Age",
               15: "Uri-Query",
               16: "Accept",
               20: "Location-Query",
               35: "Proxy-Uri",
               39: "Proxy-Scheme"
               }
    
    IF_MATCH = 1
    URI_HOST = 3
    ETAG = 4
    IF_NONE_MATCH = 5
    URI_PORT = 7
    LOCATION_PATH = 8
    URI_PATH = 11
    CONTENT_FORMAT = 12
    MAX_AGE = 14
    URI_QUERY = 15
    ACCEPT = 16
    LOCATION_QUERY = 20
    PROXY_URI = 35
    PROXY_SCHEME = 39
    
    
class COAPMessage():
    TYPES = ["CON", "NON", "ACK", "RST"]
    CON = 0
    NON = 1
    ACK = 2
    RST = 3

    def __init__(self, msg_type=0, code=0, uri=None):
        self.version = 1
        self.type    = msg_type
        self.code    = code
        self.id      = 0
        self.token   = None
        self.options = []
        self.host    = ""
        self.port    = 5683
        self.uri_path = ""
        self.content_format = None
        self.payload = None
        
        if uri != None:
            p = urlparse(uri)
            self.host = p.hostname
            if p.port:
                self.port = int(p.port)
            self.uri_path = p.path
        
    def __getOptionHeader__(self, byte):
        delta  = (byte & 0xF0) >> 4
        length = byte & 0x0F
        return (delta, length)  
        
    def __str__(self):
        result = []
        result.append("Version: %s" % self.version)
        result.append("Type: %s" % self.TYPES[self.type])
        result.append("Code: %s" % self.CODES[self.code])
        result.append("Id: %s" % self.id)
        result.append("Token: %s" % self.token)
        #result.append("Options: %s" % len(self.options))
        #for option in self.options:
        #    result.append("+ %d: %s" % (option["number"], option["value"]))
        result.append("Uri-Path: %s" % self.uri_path)
        result.append("Content-Format: %s" % (COAPContentFormat.toString(self.content_format) if self.content_format else M_PLAIN))
        result.append("Payload: %s" % self.payload)
        result.append("")
        return '\n'.join(result)
        
    def getOptionHeaderValue(self, value):
        if value > 268:
            return 14
        if value > 12:
            return 13
        return value
    
    def getOptionHeaderExtension(self, value):
        buff = bytearray()
        v = self.getOptionHeaderValue(value)
        
        if v == 14:
            value -= 269
            buff.append((value & 0xFF00) >> 8)
            buff.append(value & 0x00FF)

        elif v == 13:
            value -= 13
            buff.append(value)

        return buff
    
    def appendOption(self, buff, lastnumber, option, data):
        delta = option - lastnumber
        length = len(data)
        
        d = self.getOptionHeaderValue(delta)
        l = self.getOptionHeaderValue(length)
        
        b  = 0
        b |= (d << 4) & 0xF0  
        b |= l & 0x0F
        buff.append(b)
        
        ext = self.getOptionHeaderExtension(delta);
        for b in ext:
            buff.append(b)

        ext = self.getOptionHeaderExtension(length);
        for b in ext:
            buff.append(b)

        for b in data:
            buff.append(b)

        return option

    def getBytes(self):
        buff = bytearray()
        byte  = (self.version & 0x03) << 6
        byte |= (self.type & 0x03) << 4
        if self.token:
            token_len = min(len(self.token), 8);
        else:
            token_len = 0
        byte |= token_len
        buff.append(byte)
        buff.append(self.code)
        buff.append((self.id & 0xFF00) >> 8)
        buff.append(self.id & 0x00FF)
        
        if self.token:
            for c in self.token:
                buff.append(c)

        lastnumber = 0
        
        if len(self.uri_path) > 0:
            paths = self.uri_path.split("/")
            for p in paths:
                if len(p) > 0:
                    if PYTHON_MAJOR >= 3:
                        data = p.encode()
                    else:
                        data = bytearray(p)
                    lastnumber = self.appendOption(buff, lastnumber, COAPOption.URI_PATH, data)

        if self.content_format != None:
            data = bytearray()
            fmt_code = self.content_format
            if fmt_code > 0xFF:
                data.append((fmt_code & 0xFF00) >> 8)
            data.append(fmt_code & 0x00FF)
            lastnumber = self.appendOption(buff, lastnumber, COAPOption.CONTENT_FORMAT, data)
            
        buff.append(0xFF)
        
        if self.payload:
            if PYTHON_MAJOR >= 3:
                data = self.payload.encode()
            else:
                data = bytearray(self.payload)
            for c in data:
                buff.append(c)
        
        return buff
    
    def parseByteArray(self, buff):
        self.version = (buff[0] & 0xC0) >> 6
        self.type    = (buff[0] & 0x30) >> 4
        token_length = buff[0] & 0x0F
        index = 4
        if token_length > 0:
            self.token = buff[index:index+token_length]

        index += token_length
        self.code    = buff[1]
        self.id      = (buff[2] << 8) | buff[3]
        
        number = 0

        # process options
        while index < len(buff) and buff[index] != 0xFF:
            (delta, length) = self.__getOptionHeader__(buff[index])
            offset = 1

            # delta extended with 1 byte
            if delta == 13:
                delta += buff[index+offset]
                offset += 1
            # delta extended with 2 buff
            elif delta == 14:
                delta += 255 + ((buff[index+offset] << 8) | buff[index+offset+1])
                offset += 2
            
            # length extended with 1 byte
            if length == 13:
                length += buff[index+offset]
                offset += 1
                
            # length extended with 2 buff
            elif length == 14:
                length += 255 + ((buff[index+offset] << 8) | buff[index+offset+1])
                offset += 2

            number += delta
            valueBytes = buff[index+offset:index+offset+length]
            # opaque option value
            if number in [COAPOption.IF_MATCH, COAPOption.ETAG]:
                value = valueBytes
            # integer option value
            elif number in [COAPOption.URI_PORT, COAPOption.CONTENT_FORMAT, COAPOption.MAX_AGE, COAPOption.ACCEPT]:
                value = 0
                for b in valueBytes:
                    value <<= 8
                    value |= b
            # string option value
            else:
                if PYTHON_MAJOR >= 3:
                    value = valueBytes.decode()
                else:
                    value = str(valueBytes)
            self.options.append({'number': number, 'value': value})
            index += offset + length

        index += 1 # skip 0xFF / end-of-options
        
        if len(buff) > index:
            self.payload = buff[index:]
        else:
            self.payload = ""
        
        for option in self.options:
            (number, value) = option.values()
            if number == COAPOption.URI_PATH:
                self.uri_path += "/%s" % value


class COAPRequest(COAPMessage):
    CODES = {0: None,
             1: "GET",
             2: "POST",
             3: "PUT",
             4: "DELETE"
             }

    GET    = 1
    POST   = 2
    PUT    = 3
    DELETE = 4

    def __init__(self, msg_type=0, code=0, uri=None):
        COAPMessage.__init__(self, msg_type, code, uri)

class COAPGet(COAPRequest):
    def __init__(self, uri):
        COAPRequest.__init__(self, COAPMessage.CON, COAPRequest.GET, uri)

class COAPPost(COAPRequest):
    def __init__(self, uri):
        COAPRequest.__init__(self, COAPMessage.CON, COAPRequest.POST, uri)

class COAPPut(COAPRequest):
    def __init__(self, uri):
        COAPRequest.__init__(self, COAPMessage.CON, COAPRequest.PUT, uri)

class COAPDelete(COAPRequest):
    def __init__(self, uri):
        COAPRequest.__init__(self, COAPMessage.CON, COAPRequest.DELETE, uri)

class COAPResponse(COAPMessage):    
    CODES = {0: None,
             64: "2.00 OK",
             65: "2.01 Created",
             66: "2.02 Deleted",
             67: "2.03 Valid",
             68: "2.04 Changed",
             69: "2.05 Content",
             128: "4.00 Bad Request",
             129: "4.01 Unauthorized",
             130: "4.02 Bad Option",
             131: "4.03 Forbidden",
             132: "4.04 Not Found",
             133: "4.05 Method Not Allowed",
             134: "4.06 Not Acceptable",
             140: "4.12 Precondition Failed",
             141: "4.13 Request Entity Too Large",
             143: "4.15 Unsupported Content-Format",
             160: "5.00 Internal Server Error",
             161: "5.01 Not Implemented",
             162: "5.02 Bad Gateway",
             163: "5.03 Service Unavailable",
             164: "5.04 Gateway Timeout",
             165: "5.05 Proxying Not Supported"            
            }

    # 2.XX
    OK      = 64
    CREATED = 65
    DELETED = 66
    VALID   = 67
    CHANGED = 68
    CONTENT = 69
    
    # 4.XX
    BAD_REQUEST         = 128
    UNAUTHORIZED        = 129
    BAD_OPTION          = 130
    FORBIDDEN           = 131
    NOT_FOUND           = 132
    NOT_ALLOWED         = 133
    NOT_ACCEPTABLE      = 134
    PRECONDITION_FAILED = 140
    ENTITY_TOO_LARGE    = 141
    UNSUPPORTED_CONTENT = 143
    
    # 5.XX
    INTERNAL_ERROR          = 160
    NOT_IMPLEMENTED         = 161
    BAD_GATEWAY             = 162
    SERVICE_UNAVAILABLE     = 163
    GATEWAY_TIMEOUT         = 164
    PROXYING_NOT_SUPPORTED  = 165
    
    def __init__(self):
        COAPMessage.__init__(self)

class COAPClient():
    def __init__(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.settimeout(1.0)
        self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        
    def sendRequest(self, message):
        data = message.getBytes();
        sent = 0
        while sent<4:
            try:
                self.socket.sendto(data, (message.host, message.port))
                data = self.socket.recv(1500)
                response = COAPResponse()
                response.parseByteArray(bytearray(data))
                return response
            except socket.timeout:
                sent+=1
        return None

class COAPServer(threading.Thread):
    logger = logging.getLogger("CoAP")

    def __init__(self, host, port, handler):
        threading.Thread.__init__(self, name="COAPThread")
        self.handler = COAPHandler(handler)
        self.host = host
        self.port = port
        self.multicast_ip = '224.0.1.123'
        if socket.has_ipv6:
            address_family = socket.AF_INET6
        else:
            address_family = socket.AF_INET
        try:
            self.socket = socket.socket(address_family, socket.SOCK_DGRAM)
        except:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.bind(('', port))
        self.socket.settimeout(1)
        self.running = True
        self.start()
        
    def run(self):
        info("CoAP Server binded on coap://%s:%s/" % (self.host, self.port))
        while self.running == True:
            try:
                (request, client) = self.socket.recvfrom(1500)
                requestBytes = bytearray(request)
                coapRequest = COAPRequest()
                coapRequest.parseByteArray(requestBytes)
                coapResponse = COAPResponse()
                #self.logger.debug("Received Request:\n%s" % coapRequest)
                self.processMessage(coapRequest, coapResponse)
                #self.logger.debug("Sending Response:\n%s" % coapResponse)
                responseBytes = coapResponse.getBytes()
                self.socket.sendto(responseBytes, client)
                self.logger.debug('"%s %s CoAP/%.1f" - %s (Client: %s)' % (coapRequest.CODES[coapRequest.code], coapRequest.uri_path, coapRequest.version, coapResponse.CODES[coapResponse.code], client[0]))
                
            except socket.timeout as e:
                continue
            except Exception as e:
                if self.running == True:
                    exception(e)
            
        info("CoAP Server stopped")
    
    def enableMulticast(self):
        while not self.running:
            pass
        mreq = struct.pack("4sl", socket.inet_aton(self.multicast_ip), socket.INADDR_ANY)
        self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        info("CoAP Server binded on coap://%s:%s/ (MULTICAST)" % (self.multicast_ip, self.port))
                
    def stop(self):
        self.running = False
        self.socket.close()
        
    def processMessage(self, request, response):
        if request.type == COAPMessage.CON:
            response.type = COAPMessage.ACK
        else:
            response.type = COAPMessage.NON

        if request.token:
            response.token = request.token

        response.id = request.id
        response.uri_path = request.uri_path
        
        if request.code == COAPRequest.GET:
            self.handler.do_GET(request, response)
        elif request.code == COAPRequest.POST:
            self.handler.do_POST(request, response)
        elif request.code / 32 == 0:
            response.code = COAPResponse.NOT_IMPLEMENTED
        else:
            exception(Exception("Received CoAP Response : %s" % response))
        
class COAPHandler():
    def __init__(self, handler):
        self.handler = handler
    
    def do_GET(self, request, response):
        try:
            (code, body, contentType) = self.handler.do_GET(request.uri_path[1:], True)
            if code == 0:
                response.code = COAPResponse.NOT_FOUND
            elif code == 200:
                response.code = COAPResponse.CONTENT
            else:
                response.code =  HTTPCode2CoAPCode(code)
            response.payload = body
            response.content_format = COAPContentFormat.getCode(contentType)
        except Exception as e:
            response.code = COAPResponse.INTERNAL_ERROR
            raise e
        
    def do_POST(self, request, response):
        try:
            (code, body, contentType) = self.handler.do_POST(request.uri_path[1:], request.payload, True)
            if code == 0:
                response.code = COAPResponse.NOT_FOUND
            elif code == 200:
                response.code = COAPResponse.CHANGED
            else:
                response.code =  HTTPCode2CoAPCode(code)
            response.payload = body
            response.content_format = COAPContentFormat.getCode(contentType)
        except Exception as e:
            response.code = COAPResponse.INTERNAL_ERROR
            raise e
        





from rocket.daisy.utils.types import toint
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.spi import SPI
from rocket.daisy.devices.digital import GPIOPort

class MCP23XXX(GPIOPort):
    IODIR   = 0x00
    IPOL    = 0x01
    GPINTEN = 0x02
    DEFVAL  = 0x03
    INTCON  = 0x04
    IOCON   = 0x05
    GPPU    = 0x06
    INTF    = 0x07
    INTCAP  = 0x08
    GPIO    = 0x09
    OLAT    = 0x0A
    
    def __init__(self, channelCount):
        GPIOPort.__init__(self, channelCount)
        self.banks = int(channelCount / 8)
        
    def getAddress(self, register, channel=0):
        return register * self.banks + int(channel / 8)

    def getChannel(self, register, channel):
        self.checkDigitalChannel(channel)
        addr = self.getAddress(register, channel) 
        mask = 1 << (channel % 8)
        return (addr, mask)
    
    def __digitalRead__(self, channel):
        (addr, mask) = self.getChannel(self.GPIO, channel) 
        d = self.readRegister(addr)
        return (d & mask) == mask

    def __digitalWrite__(self, channel, value):
        (addr, mask) = self.getChannel(self.GPIO, channel) 
        d = self.readRegister(addr)
        if value:
            d |= mask
        else:
            d &= ~mask
        self.writeRegister(addr, d)
        
    def __getFunction__(self, channel):
        (addr, mask) = self.getChannel(self.IODIR, channel) 
        d = self.readRegister(addr)
        return self.IN if (d & mask) == mask else self.OUT
        
    def __setFunction__(self, channel, value):
        if not value in [self.IN, self.OUT]:
            raise ValueError("Requested function not supported")

        (addr, mask) = self.getChannel(self.IODIR, channel) 
        d = self.readRegister(addr)
        if value == self.IN:
            d |= mask
        else:
            d &= ~mask
        self.writeRegister(addr, d)

    def __portRead__(self):
        value = 0
        for i in range(self.banks):
            value |= self.readRegister(self.banks*self.GPIO+i) << 8*i
        return value
    
    def __portWrite__(self, value):
        for i in range(self.banks):
            self.writeRegister(self.banks*self.GPIO+i,  (value >> 8*i) & 0xFF)

class MCP230XX(MCP23XXX, I2C):
    def __init__(self, slave, channelCount, name):
        I2C.__init__(self, toint(slave))
        MCP23XXX.__init__(self, channelCount)
        self.name = name
        
    def __str__(self):
        return "%s(slave=0x%02X)" % (self.name, self.slave)

class MCP23008(MCP230XX):
    def __init__(self, slave=0x20):
        MCP230XX.__init__(self, slave, 8, "MCP23008")

class MCP23009(MCP230XX):
    def __init__(self, slave=0x20):
        MCP230XX.__init__(self, slave, 8, "MCP23009")

class MCP23017(MCP230XX):
    def __init__(self, slave=0x20):
        MCP230XX.__init__(self, slave, 16, "MCP23017")

class MCP23018(MCP230XX):
    def __init__(self, slave=0x20):
        MCP230XX.__init__(self, slave, 16, "MCP23018")

class MCP23SXX(MCP23XXX, SPI):
    SLAVE = 0x20
    
    WRITE = 0x00
    READ  = 0x01
    
    def __init__(self, chip, slave, channelCount, name):
        SPI.__init__(self, toint(chip), 0, 8, 10000000)
        MCP23XXX.__init__(self, channelCount)
        self.slave = self.SLAVE
        iocon_value = 0x08 # Hardware Address Enable
        iocon_addr  = self.getAddress(self.IOCON)
        self.writeRegister(iocon_addr, iocon_value)
        self.slave = toint(slave)
        self.name = name

    def __str__(self):
        return "%s(chip=%d, slave=0x%02X)" % (self.name, self.chip, self.slave)

    def readRegister(self, addr):
        d = self.xfer([(self.slave << 1) | self.READ, addr, 0x00])
        return d[2]

    def writeRegister(self, addr, value):
        self.writeBytes([(self.slave << 1) | self.WRITE, addr, value])
    
class MCP23S08(MCP23SXX):
    def __init__(self, chip=0, slave=0x20):
        MCP23SXX.__init__(self, chip, slave, 8, "MCP23S08")

class MCP23S09(MCP23SXX):
    def __init__(self, chip=0, slave=0x20):
        MCP23SXX.__init__(self, chip, slave, 8, "MCP23S09")

class MCP23S17(MCP23SXX):
    def __init__(self, chip=0, slave=0x20):
        MCP23SXX.__init__(self, chip, slave, 16, "MCP23S17")

class MCP23S18(MCP23SXX):
    def __init__(self, chip=0, slave=0x20):
        MCP23SXX.__init__(self, chip, slave, 16, "MCP23S18")




# +-------------------------------------------------------------------------------+
#
#      Program:    setup.py
#
#      Purpose:    setup for remote open control - key enabling technology (Rocket)
#
#      Target:     Maxwell SOC on isar
#
#      Author:     Martin Shishkov
#
#      License:    GPL 3
# +-------------------------------------------------------------------------------+
import os

class Init:
    def __init__(self):
        print('Installing rocket-daisy')
        
    def get_package(self):
        linkPyPi = "https://files.pythonhosted.org/packages/14/02/760b94aa576091428d1bf3d25cacbc03d0e771e255d3f98340652668b2cb/rocket-daisy-0.1a8.tar.gz"
        rocket = "rocket-daisy-0.1a8"
        if not os.path.isdir(rocket):
            print('downloading rocket-daisy')
            os.system("wget " + linkPyPi)
            os.system(("tar xvzf %s.tar.gz" % (rocket)))
            os.chdir(rocket)
            os.system("sudo python3 setup.py bdist")
            
            
 
 
 
i = Init()
i.get_package()    



            
import time
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.sensor import Luminosity, Distance
from rocket.daisy.utils.types import toint
from rocket.daisy.utils.logger import debug 

class VCNL4000(I2C, Luminosity, Distance):
    REG_COMMAND           = 0x80
    REG_IR_LED_CURRENT    = 0x83
    REG_AMB_PARAMETERS    = 0x84
    REG_AMB_RESULT_HIGH   = 0x85
    REG_PROX_RESULT_HIGH  = 0x87
    REG_PROX_FREQUENCY    = 0x89
    REG_PROX_ADJUST       = 0x8A
   
    VAL_MOD_TIMING_DEF    = 129 # default from data sheet
    
    VAL_PR_FREQ_3M125HZ   = 0
    VAL_PR_FREQ_1M5625HZ  = 1
    VAL_PR_FREQ_781K25HZ  = 2
    VAL_PR_FREQ_390K625HZ = 3
    
    VAL_START_AMB         = 1 << 4
    VAL_START_PROX        = 1 << 3
    
    VAL_INVALID           = -1
    VAL_NO_PROXIMITY      = -1
    
    MASK_PROX_FREQUENCY  = 0b00111111
    MASK_IR_LED_CURRENT  = 0b00111111
    MASK_PROX_READY      = 0b00100000
    MASK_AMB_READY       = 0b01000000
    
    def __init__(self, slave=0b0010011, current=20, frequency=781, prox_threshold=15, prox_cycles=10, cal_cycles= 5):
        I2C.__init__(self, toint(slave))
        self.setCurrent(toint(current))
        self.setFrequency(toint(frequency))
        self.prox_threshold = toint(prox_threshold)
        self.prox_cycles = toint(prox_cycles)
        self.cal_cycles = toint(cal_cycles)
        self.__setProximityTiming__()
        self.__setAmbientMeasuringMode__()
        time.sleep(0.001)
        self.calibrate() # may have to be repeated from time to time or before every proximity measurement

    def __str__(self):
        return "VCNL4000(slave=0x%02X)" % self.slave

    def __family__(self):
        return [Luminosity.__family__(self), Distance.__family__(self)]

    def __setProximityTiming__(self):
        self.writeRegister(self.REG_PROX_ADJUST, self.VAL_MOD_TIMING_DEF)
                               
    def __setAmbientMeasuringMode__(self):
        ambient_parameter_bytes = 1 << 7 | 1 << 3 | 5
        # Parameter is set to
        # -continuous conversion mode (bit 7)
        # -auto offset compensation (bit 3)
        # -averaging 32 samples (5)
        self.writeRegister(self.REG_AMB_PARAMETERS, ambient_parameter_bytes)

    def calibrate(self):
        self.offset = self.__measureOffset__()
        debug ("VCNL4000: offset = %d" % (self.offset))
        return self.offset

        
    def setCurrent(self, current):
        self.current = current
        self.__setCurrent__()
        

    def getCurrent(self):
        return self.__getCurrent__()

    def setFrequency(self, frequency):
        self.frequency = frequency
        self.__setFrequency__()

    def getFrequency(self):
        return self.__getFrequency__()

    def __setFrequency__(self):
        if not self.frequency in [391, 781, 1563, 3125]:
            raise ValueError("Frequency %d out of range [%d,%d,%d,,%d]" % (self.frequency, 391, 781, 1563, 3125))
        if self.frequency == 391:
            bits_frequency = self.VAL_PR_FREQ_390K625HZ
        elif self.frequency == 781:
            bits_frequency = self.VAL_PR_FREQ_781K25HZ
        elif self.frequency == 1563:
            bits_frequency = self.VAL_PR_FREQ_1M5625HZ            
        elif self.frequency == 3125:
            bits_frequency = self.VAL_PR_FREQ_3M125HZ
        self.writeRegister(self.REG_PROX_FREQUENCY, bits_frequency)
        debug ("VCNL4000: new freq = %d" % (self.readRegister(self.REG_PROX_FREQUENCY)))
        
    def __getFrequency__(self):
        bits_frequency = self.readRegister(self.REG_PROX_FREQUENCY) & self.MASK_PROX_FREQUENCY
        if bits_frequency == self.VAL_PR_FREQ_390K625HZ:
            f =  391
        elif bits_frequency == self.VAL_PR_FREQ_781K25HZ:
            f =  781
        elif bits_frequency == self.VAL_PR_FREQ_1M5625HZ:
            f = 1563
        elif bits_frequency == self.VAL_PR_FREQ_3M125HZ:
            f = 3125
        else:
            f = self.VAL_INVALID # indicates undefined
        return f

    def __setCurrent__(self):
        if not self.current in range(0,201):
            raise ValueError("%d mA LED current out of range [%d..%d] mA" % (self.current, 0, 201))
        self.writeRegister(self.REG_IR_LED_CURRENT, int(self.current / 10))       
        debug ("VCNL4000: new curr = %d" % (self.readRegister(self.REG_IR_LED_CURRENT)))
        
    def __getCurrent__(self):
        bits_current = self.readRegister(self.REG_IR_LED_CURRENT) & self.MASK_IR_LED_CURRENT
        return bits_current * 10
        
    def __getLux__(self):
        self.writeRegister(self.REG_COMMAND, self.VAL_START_AMB)
        while not (self.readRegister(self.REG_COMMAND) & self.MASK_AMB_READY):
            time.sleep(0.001)
        light_bytes = self.readRegisters(self.REG_AMB_RESULT_HIGH, 2)
        light_word = light_bytes[0] << 8 | light_bytes[1]
        return self.__calculateLux__(light_word)
         
    def __calculateLux__(self, light_word):
        return (light_word + 3) * 0.25 # From VISHAY application note

    def __getMillimeter__(self):
        success = 0
        fail = 0
        prox = 0
        match_cycles = self.prox_cycles
        while (fail < match_cycles) & (success < match_cycles):
            real_counts = self.__readProximityCounts__() - self.offset 
            if real_counts > self.prox_threshold:
                success += 1
                prox += real_counts
            else:
                fail += 1
        if fail == match_cycles:
            return self.VAL_NO_PROXIMITY
        else:
            return self.__calculateMillimeter__(prox // match_cycles)
    
    def __calculateMillimeter__(self, raw_proximity_counts):
        # According to chip spec the proximity counts are strong non-linear with distance and cannot be calculated
        # with a direct formula. From experience found on web this chip is generally not suited for really exact
        # distance calculations. This is a rough distance estimation lookup table for now. Maybe someone can
        # provide a more exact approximation in the future.

        debug ("VCNL4000: prox real raw counts = %d" % (raw_proximity_counts))
        if raw_proximity_counts >= 10000:
            estimated_distance = 0
        elif raw_proximity_counts >= 3000:
            estimated_distance = 5
        elif raw_proximity_counts >= 900:
            estimated_distance = 10
        elif raw_proximity_counts >= 300:
            estimated_distance = 20
        elif raw_proximity_counts >= 150:
            estimated_distance = 30
        elif raw_proximity_counts >= 75:
            estimated_distance = 40
        elif raw_proximity_counts >= 50:
            estimated_distance = 50
        elif raw_proximity_counts >= 25:
            estimated_distance = 70
        else:
            estimated_distance = 100
        return estimated_distance
    
    def __measureOffset__(self):
        offset = 0
        for unused in range(self.cal_cycles):
            offset += self.__readProximityCounts__()
        return offset // self.cal_cycles
        
    def __readProximityCounts__(self):
        self.writeRegister(self.REG_COMMAND, self.VAL_START_PROX)
        while not (self.readRegister(self.REG_COMMAND) & self.MASK_PROX_READY):
            time.sleep(0.001)
        proximity_bytes = self.readRegisters(self.REG_PROX_RESULT_HIGH, 2)
        debug ("VCNL4000: prox raw value = %d" % (proximity_bytes[0] << 8 | proximity_bytes[1]))
        return (proximity_bytes[0] << 8 | proximity_bytes[1])
    





from rocket.daisy.utils.types import toint
from rocket.daisy.utils.types import M_JSON
from rocket.daisy.devices.instance import deviceInstance
from rocket.daisy.decorators.rest import request, response

class Pressure():
    def __init__(self, altitude=0, external=None):
        self.altitude = toint(altitude)
        if isinstance(external, str):
            self.external = deviceInstance(external)
        else:
            self.external = external
        
        if self.external != None and not isinstance(self.external, Temperature):
            raise Exception("external must be a Temperature sensor")

    def __family__(self):
        return "Pressure"

    def __getPascal__(self):
        raise NotImplementedError
    
    def __getPascalAtSea__(self):
        raise NotImplementedError
    
    @request("GET", "sensor/pressure/pa")
    @response("%d")
    def getPascal(self):
        return self.__getPascal__()
    
    @request("GET", "sensor/pressure/hpa")
    @response("%.2f")
    def getHectoPascal(self):
        return float(self.__getPascal__()) / 100.0
    
    @request("GET", "sensor/pressure/sea/pa")
    @response("%d")
    def getPascalAtSea(self):
        pressure = self.__getPascal__()
        if self.external != None:
            k = self.external.getKelvin()
            if k != 0:
                return float(pressure) / (1.0 / (1.0 + 0.0065 / k * self.altitude)**5.255)
        return float(pressure) / (1.0 - self.altitude / 44330.0)**5.255

    @request("GET", "sensor/pressure/sea/hpa")
    @response("%.2f")
    def getHectoPascalAtSea(self):
        return self.getPascalAtSea() / 100.0
    
class Temperature():
    def __family__(self):
        return "Temperature"
    
    def __getKelvin__(self):
        raise NotImplementedError
    
    def __getCelsius__(self):
        raise NotImplementedError

    def __getFahrenheit__(self):
        raise NotImplementedError
    
    def Kelvin2Celsius(self, value=None):
        if value == None:
            value = self.getKelvin()
        return value - 273.15
    
    def Kelvin2Fahrenheit(self, value=None):
        if value == None:
            value = self.getKelvin()
        return value * 1.8 - 459.67
    
    def Celsius2Kelvin(self, value=None):
        if value == None:
            value = self.getCelsius()
        return value + 273.15
    
    def Celsius2Fahrenheit(self, value=None):
        if value == None:
            value = self.getCelsius()
        return value * 1.8 + 32

    def Fahrenheit2Kelvin(self, value=None):
        if value == None:
            value = self.getFahrenheit()
        return (value - 459.67) / 1.8

    def Fahrenheit2Celsius(self, value=None):
        if value == None:
            value = self.getFahrenheit()
        return (value - 32) / 1.8

    @request("GET", "sensor/temperature/k")
    @response("%.02f")
    def getKelvin(self):
        return self.__getKelvin__()
    
    @request("GET", "sensor/temperature/c")
    @response("%.02f")
    def getCelsius(self):
        return self.__getCelsius__()
    
    @request("GET", "sensor/temperature/f")
    @response("%.02f")
    def getFahrenheit(self):
        return self.__getFahrenheit__()
    
class Luminosity():
    def __family__(self):
        return "Luminosity"
    
    def __getLux__(self):
        raise NotImplementedError

    @request("GET", "sensor/luminosity/lux")
    @response("%.02f")
    def getLux(self):
        return self.__getLux__()

class Distance():
    def __family__(self):
        return "Distance"
    
    def __getMillimeter__(self):
        raise NotImplementedError

    @request("GET", "sensor/distance/mm")
    @response("%.02f")
    def getMillimeter(self):
        return self.__getMillimeter__()
    
    @request("GET", "sensor/distance/cm")
    @response("%.02f")
    def getCentimeter(self):
        return self.getMillimeter() / 10
    
    @request("GET", "sensor/distance/m")
    @response("%.02f")
    def getMeter(self):
        return self.getMillimeter() / 1000
    
    @request("GET", "sensor/distance/in")
    @response("%.02f")
    def getInch(self):
        return self.getMillimeter() / 0.254
    
    @request("GET", "sensor/distance/ft")
    @response("%.02f")
    def getFoot(self):
        return self.getInch() / 12
    
    @request("GET", "sensor/distance/yd")
    @response("%.02f")
    def getYard(self):
        return self.getInch() / 36
    
class Humidity():
    def __family__(self):
        return "Humidity"
    
    def __getHumidity__(self):
        raise NotImplementedError    
    
    @request("GET", "sensor/humidity/float")
    @response("%f")
    def getHumidity(self):
        return self.__getHumidity__()
    
    @request("GET", "sensor/humidity/percent")
    @response("%d")
    def getHumidityPercent(self):
        return self.__getHumidity__() * 100
        
DRIVERS = {}
DRIVERS["bmp085"] = ["BMP085", "BMP180"]
DRIVERS["onewiretemp"] = ["DS1822", "DS1825", "DS18B20", "DS18S20", "DS28EA00"]
DRIVERS["tmpXXX"] = ["TMP75", "TMP102", "TMP275"]
DRIVERS["tslXXXX"] = ["TSL2561", "TSL2561CS", "TSL2561T", "TSL4531", "TSL45311", "TSL45313", "TSL45315", "TSL45317"]
DRIVERS["vcnl4000"] = ["VCNL4000"]
DRIVERS["hytXXX"] = ["HYT221"]






from rocket.daisy.utils.types import toint
from rocket.daisy.devices.spi import SPI
from rocket.daisy.devices.analog import DAC

class MCP492X(SPI, DAC):
    def __init__(self, chip, channelCount, vref):
        SPI.__init__(self, toint(chip), 0, 8, 10000000)
        DAC.__init__(self, channelCount, 12, float(vref))
        self.buffered=False
        self.gain=False
        self.shutdown=False
        self.values = [0 for i in range(channelCount)]

    def __str__(self):
        return "MCP492%d(chip=%d)" % (self._analogCount, self.chip)

    def __analogRead__(self, channel, diff=False):
        return self.values[channel]
        
    def __analogWrite__(self, channel, value):
        d = bytearray(2)
        d[0]  = 0
        d[0] |= (channel & 0x01) << 7
        d[0] |= (self.buffered & 0x01) << 6
        d[0] |= (not self.gain & 0x01) << 5
        d[0] |= (not self.shutdown & 0x01) << 4
        d[0] |= (value >> 8) & 0x0F
        d[1]  = value & 0xFF
        self.writeBytes(d)
        self.values[channel] = value

class MCP4921(MCP492X):
    def __init__(self, chip=0, vref=3.3):
        MCP492X.__init__(self, chip, 1)

class MCP4922(MCP492X):
    def __init__(self, chip=0, vref=3.3):
        MCP492X.__init__(self, chip, 2)




import json
from rocket.daisy.utils import logger

M_PLAIN = "text/plain"
M_JSON  = "application/json"

def jsonDumps(obj):
    if logger.debugEnabled():
        return json.dumps(obj, sort_keys=True, indent=4, separators=(',', ': '))
    else:
        return json.dumps(obj)
    
def str2bool(value):
    return (value == "1") or (value == "true") or (value == "True") or (value == "yes") or (value == "Yes")

def toint(value):
    if isinstance(value, str):
        if value.startswith("0b"):
            return int(value, 2)
        elif value.startswith("0x"):
            return int(value, 16)
        else:
            return int(value)
    return value

        
def signInteger(value, bitcount):
    if value & (1<<(bitcount-1)):
        return value - (1<<bitcount)
    return value





import os
import time
import subprocess

from rocket.daisy.utils.logger import debug, info

BUSLIST = {
    "I2C": {"enabled": False, "gpio": {0:"SDA", 1:"SCL", 2:"SDA", 3:"SCL"}, "modules": ["i2c-bcm2708", "i2c-dev"]},
    "SPI": {"enabled": False, "gpio": {7:"CE1", 8:"CE0", 9:"MISO", 10:"MOSI", 11:"SCLK"}, "modules": ["spi-bcm2708", "spidev"]},
    "UART": {"enabled": False, "gpio": {14:"TX", 15:"RX"}},
    "ONEWIRE": {"enabled": False, "gpio": {4:"DATA"}, "modules": ["w1-gpio"], "wait": 2}
}

def loadModule(module):
    debug("Loading module : %s" % module)
    subprocess.call(["modprobe", module])
    
def unloadModule(module):
    subprocess.call(["modprobe", "-r", module])
    
def loadModules(bus):
    if BUSLIST[bus]["enabled"] == False and not modulesLoaded(bus):
        info("Loading %s modules" % bus)
        for module in BUSLIST[bus]["modules"]:
            loadModule(module)
        if "wait" in BUSLIST[bus]:
            info("Sleeping %ds to let %s modules load" % (BUSLIST[bus]["wait"], bus))
            time.sleep(BUSLIST[bus]["wait"])

    BUSLIST[bus]["enabled"] = True

def unloadModules(bus):
    info("Unloading %s modules" % bus)
    for module in BUSLIST[bus]["modules"]:
        unloadModule(module)
    BUSLIST[bus]["enabled"] = False
        
def __modulesLoaded__(modules, lines):
    if len(modules) == 0:
        return True
    for line in lines:
        if modules[0].replace("-", "_") == line.split(" ")[0]:
            return __modulesLoaded__(modules[1:], lines)
    return False

def modulesLoaded(bus):
    if not bus in BUSLIST or not "modules" in BUSLIST[bus]:
        return True

    try:
        with open("/proc/modules") as f:
            c = f.read()
            f.close()
            lines = c.split("\n")
            return __modulesLoaded__(BUSLIST[bus]["modules"], lines)
    except:
        return False

def checkAllBus():
    for bus in BUSLIST:
        if modulesLoaded(bus):
            BUSLIST[bus]["enabled"] = True

class Bus():
    def __init__(self, busName, device, flag=os.O_RDWR):
        loadModules(busName)
        self.busName = busName
        self.device = device
        self.flag = flag
        self.fd = 0
        self.open()
        
    def open(self):
        self.fd = os.open(self.device, self.flag)
        if self.fd < 0:
            raise Exception("Cannot open %s" % self.device)

    def close(self):
        if self.fd > 0:
            os.close(self.fd)
    
    def read(self, size=1):
        if self.fd > 0:
            return os.read(self.fd, size)
        raise Exception("Device %s not open" % self.device)
    
    def readBytes(self, size=1):
        return bytearray(self.read(size))
    
    def readByte(self):
        return self.readBytes()[0]

    def write(self, string):
        if self.fd > 0:
            return os.write(self.fd, string)
        raise Exception("Device %s not open" % self.device)
    
    def writeBytes(self, data):
        return self.write(bytearray(data))

    def writeByte(self, value):
        self.writeBytes([value])
        



import imp
from rocket.daisy.utils.logger import info, exception, debug
import rocket.daisy.utils.thread as thread
SCRIPTS = {}

def loadScript(name, source, handler = None):
    info("Loading %s from %s" % (name, source))
    script = imp.load_source(name, source)
    SCRIPTS[name] = script

    if hasattr(script, "setup"):
        script.setup()
    if handler:
        for aname in dir(script):
            attr = getattr(script, aname)
            if callable(attr) and hasattr(attr, "macro"):
                debug("### daisy, gulliversoft, loadScript %s handler %s" % (aname, handler))
                handler.addMacro(attr)
            else:
                debug("### daisy, gulliversoft, loadScript %s is not callable" % (aname))
    else:
        debug("### daisy, gulliversoft, loadScript called without handler")

    if hasattr(script, "loop"):
        thread.runLoop(script.loop, True)

def unloadScripts():
    for name in SCRIPTS:
        script = SCRIPTS[name]
        if hasattr(script, "destroy"):
            script.destroy()
    




import os
import fcntl
import struct
import termios

from rocket.daisy.devices.bus import Bus
from rocket.daisy.decorators.rest import request, response

TIOCINQ   = hasattr(termios, 'FIONREAD') and termios.FIONREAD or 0x541B
TIOCM_zero_str = struct.pack('I', 0)

class Serial(Bus):
    def __init__(self, device="/dev/ttyAMA0", baudrate=9600):
        if not device.startswith("/dev/"):
            device = "/dev/%s" % device
        
        if isinstance(baudrate, str):
            baudrate = int(baudrate)

        aname = "B%d" % baudrate
        if not hasattr(termios, aname):
            raise Exception("Unsupported baudrate")
        self.baudrate = baudrate

        Bus.__init__(self, "UART", device, os.O_RDWR | os.O_NOCTTY)
        fcntl.fcntl(self.fd, fcntl.F_SETFL, os.O_NDELAY)
        
        #backup  = termios.tcgetattr(self.fd)
        options = termios.tcgetattr(self.fd)
        # iflag
        options[0] = 0

        # oflag
        options[1] = 0

        # cflag
        options[2] |= (termios.CLOCAL | termios.CREAD)
        options[2] &= ~termios.PARENB
        options[2] &= ~termios.CSTOPB
        options[2] &= ~termios.CSIZE
        options[2] |= termios.CS8

        # lflag
        options[3] = 0

        speed = getattr(termios, aname)
        # input speed
        options[4] = speed
        # output speed
        options[5] = speed
        
        termios.tcsetattr(self.fd, termios.TCSADRAIN, options)
        
    def __str__(self):
        return "Serial(%s, %dbps)" % (self.device, self.baudrate)
    
    def __family__(self):
        return "Serial"
    
    def available(self):
        s = fcntl.ioctl(self.fd, TIOCINQ, TIOCM_zero_str)
        return struct.unpack('I',s)[0]
    
    @request("GET", "")
    @response("%s")
    def readString(self):
        if self.available() > 0:
            return self.read(self.available()).decode()
        return ""
    
    @request("POST", "", "data")
    def writeString(self, data):
        if isinstance(data, str):
            self.write(data.encode())
        else:
            self.write(data)




from time import sleep
from rocket.daisy.utils.types import toint
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.sensor import Temperature,Humidity

class HYT221(I2C, Temperature, Humidity):
    VAL_RETRIES = 30
    
    def __init__(self, slave=0x28):
        I2C.__init__(self, toint(slave))
        self.__startMeasuring__()
        
    def __str__(self):
        return "HYT221(slave=0x%02X)" % self.slave
            
    def __family__(self):
        return [Temperature.__family__(self), Humidity.__family__(self)]
    
    def __startMeasuring__(self):
      self.writeByte(0x0)
      
    def readRawData(self):
        self.__startMeasuring__()
        for i in range(self.VAL_RETRIES):
            #C-code example from sensor manufacturer suggest to wait 100ms (Duration of the measurement)
            # no to get the very last measurement shoudn't be a problem -> wait 10ms
            # try a read every 10 ms for maximum VAL_RETRIES times
            sleep(.01)
            data_bytes=self.readBytes(4)
            stale_bit = (data_bytes[0] & 0b01000000) >> 6
            if (stale_bit == 0):    
                raw_t = ((data_bytes[2] << 8) | data_bytes[3]) >> 2
                raw_h = ((data_bytes[0] & 0b00111111) << 8) | data_bytes[1]
                return (raw_t, raw_h)

        #Stale was never 0, so datas are not actual
        raise Exception("HYT221(slave=0x%02X): data fetch timeout" % self.slave)
            
    
    def __getCelsius__(self):
        (raw_t, raw_h) = self.readRawData()
        if raw_t < 0x3FFF:
            return (raw_t * 165.0 / 2**14) - 40.0
        else:
            raise ValueError("Temperature value out of range (RawValue=0x%04X Max:0x3FFF)" % raw_t)
    
    def __getFahrenheit__(self):
        return self.Celsius2Fahrenheit()

    def __getKelvin__(self):
        return self.Celsius2Kelvin()

    def __getHumidity__(self):
        (raw_t, raw_h) = self.readRawData()
        if raw_h < 0x3FFF:
            return raw_h * 1.0 / 2**14
        else:
            raise ValueError("Humidity value out of range (RawValue=0x%04X Max:0x3FFF)" % raw_h)
       



# +---------------------------------------------------------------------------+
#
#      Program:    maxmotion.py
#
#      Purpose:    A daisy backend for the RC skidloader				
#      
#      Target:     SITARA
#
#      Author:     Martin Shishkov
#
#      License:    GPL 3
# +---------------------------------------------------------------------------+

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'modules'))

import direction_converter

if os.name == 'nt':
    sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'standalone'))
    import mock_dataexchange as deb
    import mock_daisy as daisy
else:
    import daisy
    import _DataExchangeBuffer as deb

SIZE = 15
OK = 0

scriptpath = os.path.dirname(os.path.realpath(__file__))
command = None


 
# setup function is automatically called at WebIOPi startup
def setup():
    assert deb.connect() == OK

# loop function is repeatedly called by Daisy
def loop():
	daisy.sleep(1)
	global command
	if command is not None:	
		print ("Command:----------------- SENT --------------------------")
		print (command)
		print ("-----------------------------------------------------------")


# destroy function is called at Daisy shutdown
def destroy():
	deb.disconnect()


@daisy.macro
def Forward():
	global command
	global SIZE
	command = Combine(command, direction_converter.directionToArray("Ahead"))
	assert deb.send(command, SIZE) == OK

@daisy.macro
def TurnLeft():
	global command
	global SIZE
	command = Combine(command, direction_converter.directionToArray("Left"))
	assert deb.send(command, SIZE) == OK

@daisy.macro
def Reverse():
	global command
	global SIZE
	command = Combine(command, direction_converter.directionToArray("Backward"))
	assert deb.send(command, SIZE) == OK

@daisy.macro
def TurnRight():
	global command
	global SIZE
	command = Combine(command, direction_converter.directionToArray("Right"))
	assert deb.send(command, SIZE) == OK

@daisy.macro
def Stop():
	global command
	global SIZE
	command = direction_converter.directionToArray("Stop")
	assert deb.send(command, SIZE) == OK
	

@daisy.macro
def ArmUp():
	global command
	global SIZE
	command = Combine(command, direction_converter.directionToArray("ArmUp"))
	assert deb.send(command, SIZE) == OK
	
@daisy.macro
def ArmDown():
	global command
	global SIZE
	command = direction_converter.directionToArray("ArmDown")
	assert deb.send(command, SIZE) == OK
	
@daisy.macro
def TiltUp():
	global command
	global SIZE
	command = Combine(command, direction_converter.directionToArray("TiltUp"))
	assert deb.send(command, SIZE) == OK
	
	
@daisy.macro
def TiltDown():
	global command
	global SIZE
	command = direction_converter.directionToArray("TiltDown")
	assert deb.send(command, SIZE) == OK
	
@daisy.macro
def Lights():
	global command
	global SIZE
	command = Combine(command, direction_converter.directionToArray("Lights"))
	assert deb.send(command, SIZE) == OK

@daisy.macro
def FlashAll():
	global command
	global SIZE
	command = direction_converter.directionToArray("Parked")
	assert deb.send(command, SIZE) == OK
	daisy.sleep(1)
	command = Combine(command,direction_converter.directionToArray("DI"))
	assert deb.send(command, SIZE) == OK
	daisy.sleep(1)
	command = Combine(command, direction_converter.directionToArray("GG"))
	assert deb.send(command, SIZE) == OK
	daisy.sleep(1)
	command = Combine(command,direction_converter.directionToArray("ER"))
	assert deb.send(command, SIZE) == OK

@daisy.macro	
def Move(L, R):
	global command
	global SIZE
	command = direction_converter.directionToJoystick("LR", L, R)
	assert deb.send(command, SIZE) == OK

#Parameter                  Byte number
DIRECTION_LEFT_WHEELS       = 0
SPEED_LEFT_WHEELS           = 1
DIRECTION_RIGHT_WHEELS      = 3
SPEED_RIGHT_WHEELS          = 4
ARM_POS                     = 6
TILT_POS                    = 9

def Combine(cmd1 :list, cmd2 :list):
        sum = []
        for x in range(0, len(cmd1), 1):
            sum.append(int(cmd1[x], 16) + int(cmd2[x], 16))

        sum[SPEED_LEFT_WHEELS] %= 0xFF
        sum[SPEED_RIGHT_WHEELS] %= 0xFF
        sum[ARM_POS] %= (0xC - 0x4)
        sum[TILT_POS] %= (0xC - 0x4)
        sequence = ("".join(hex(byte & 0x0F) for byte in sum))
        sequence = sequence.replace("0x", "")
        sequence = sequence.upper()
        return sequence




	#Byte number	Parameter
	#0	Drive direction left wheels, 0=forward, 1=backwards
	#1	Drive speed left wheels, range 0..255, 0=stop, 255=full speed (1st nibble)
	#2	Drive speed left wheels, range 0..255, 0=stop, 255=full speed (2nd nibble)
	#3	Drive direction right wheels, 0=forward, 1=backwards
	#4	Drive speed right wheels, range 0..255, 0=stop, 255=full speed (1st nibble)
	#5	Drive speed right wheels, range 0..255, 0=stop, 255=full speed (2nd nibble)
	#6	Shovel arm position, range 1382..2764 (1st nibble), details on handling tbd
	#7	Shovel arm position, range 1382..2764 (2st nibble), details on handling tbd
	#8	Shovel arm position, range 1382..2764 (3st nibble), details on handling tbd
	#9	Shovel tilting position, range 1382..2764 (1st nibble), details on handling tbd
	#10	Shovel tilting position, range 1382..2764 (2st nibble), details on handling tbd
	#11	Shovel tilting position, range 1382..2764 (3st nibble), details on handling tbd
	#12	Reserved for actor control (lights, camera on/on, sound module, etc.) (1st nibble)
	#13	Reserved for actor control (lights, camera on/on, sound module, etc.) (2st nibble)
	#14	CR (carriage return), ASCII 13, as telegram termination flag



#     ____        _           
#    / __ \____ _(_)______  __
#   / / / / __ `/ / ___/ / / /
#  / /_/ / /_/ / (__  ) /_/ / 
# /_____/\__,_/_/____/\__, /  
# martin shishkov    /____/   
#
#   Copyright 2020 Martin Shishkov, gulliversoft.com
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import time
import socket

from rocket.daisy.utils.config import Config
from rocket.daisy.utils import loader

from rocket.daisy.utils import crypto
from rocket.daisy.devices import manager
from rocket.daisy.protocols import rest
from rocket.daisy.protocols import http
from rocket.daisy.protocols import coap
from rocket.daisy.devices.digital.gpio import NativeGPIO
from rocket.daisy.utils.logger import info, exception, debug, warn

def getLocalIP():
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect(('8.8.8.8', 53))
            host = s.getsockname()[0]
            s.close()
            return host 
        except:
            return "localhost"

class Server():
    def __init__(self, port=8000, coap_port=5683, login=None, password=None, passwdfile=None, configfile=None, scriptfile=None):
        self.host = getLocalIP()
        self.gpio = NativeGPIO()
        self.restHandler = rest.RESTHandler()
        manager.addDeviceInstance("GPIO", self.gpio, [])

        if configfile != None:
            info("Loading configuration from %s" % configfile)
            config = Config(configfile)
        else:
            config = Config()
            
        self.gpio.addSetups(config.items("GPIO"))
        self.gpio.addResets(config.items("~GPIO"))
        self.gpio.setup()
        
        devices = config.items("DEVICES")
        for (name, params) in devices:
            values = params.split(" ")
            driver = values[0];
            args = {}
            i = 1
            while i < len(values):
                (arg, val) = values[i].split(":")
                args[arg] = val
                i+=1
            manager.addDevice(name, driver, args)
        

        if scriptfile != None:
            scriptname = scriptfile.split("/")[-1].split(".")[0]
            loader.loadScript(scriptname, scriptfile, self.restHandler)    
        
        scripts = config.items("SCRIPTS")
        for (name, source) in scripts:
            debug("### daisy, evosoft, loader is getting source: %s" % source)
            loader.loadScript(name, source, self.restHandler)
        
        self.restHandler.device_mapping = config.getboolean("REST", "device-mapping", True)
        self.gpio.post_value = config.getboolean("REST", "gpio-post-value", True)
        self.gpio.post_function = config.getboolean("REST", "gpio-post-function", True)
        exports = config.get("REST", "gpio-export", None)
        if exports != None:
            self.gpio.export = [int(s) for s in exports.split(",")]
        self.restHandler.export = self.gpio.export
        
        http_port = config.getint("HTTP", "port", port)
        http_enabled = config.getboolean("HTTP", "enabled", http_port > 0)
        http_passwdfile = config.get("HTTP", "passwd-file", passwdfile)
        context = config.get("HTTP", "context", None)
        docroot = config.get("HTTP", "doc-root", None)
        index = config.get("HTTP", "welcome-file", None)
            
        coap_port = config.getint("COAP", "port", coap_port)
        coap_enabled = config.getboolean("COAP", "enabled", coap_port > 0)
        coap_multicast = config.getboolean("COAP", "multicast", coap_enabled)

        routes = config.items("ROUTES")
        for (source, destination) in routes:
            self.restHandler.addRoute(source, destination)
    
        auth = None
        if http_passwdfile != None:
            if os.path.exists(http_passwdfile):
                f = open(http_passwdfile)
                auth = f.read().strip(" \r\n")
                f.close()
                if len(auth) > 0:
                    info("Access protected using %s" % http_passwdfile)
                else:
                    info("Passwd file %s is empty" % http_passwdfile)
            else:
                error("Passwd file %s not found" % http_passwdfile)
            
        elif login != None or password != None:
            auth = crypto.encryptCredentials(login, password)
            info("Access protected using login/password")
            
        if auth == None or len(auth) == 0:
            warn("Access unprotected")
        
        realm = config.get("HTTP", "prompt", None)
        
        if http_enabled:
            self.http_server = http.HTTPServer(self.host, http_port, self.restHandler, context, docroot, index, auth, realm)
        else:
            self.http_server = None
        
        if coap_enabled:
            self.coap_server = coap.COAPServer(self.host, coap_port, self.restHandler)
            if coap_multicast:
                self.coap_server.enableMulticast()
        else:
            self.coap_server = None
    
    def addMacro(self, macro):
        debug("### daisy, gulliversoft, addMacro %s" % macro)
        self.restHandler.addMacro(macro)
        
    def stop(self):
        if self.http_server:
            self.http_server.stop()
        if self.coap_server:
            self.coap_server.stop()
        loader.unloadScripts()
        manager.closeDevices()







import sys
from rocket.daisy.devices.onewire import OneWire
from rocket.daisy.devices.sensor import Temperature

class OneWireTemp(OneWire, Temperature):
    def __init__(self, slave=None, family=0, name="1-Wire"):
        OneWire.__init__(self, slave, family, "TEMP")
        self.name = name
        
    def __str__(self):
        return "%s(slave=%s)" % (self.name, self.slave)
    
    def __getKelvin__(self):
        return self.Celsius2Kelvin()

    def __getCelsius__(self):
        data = self.read()
        lines = data.split("\n")
        if lines[0].endswith("YES"):
            i = lines[1].find("=")
            temp = lines[1][i+1:]
            return int(temp) / 1000.0
        return (-sys.maxsize - 1) / 1000.0
    
    def __getFahrenheit__(self):
        return self.Celsius2Fahrenheit()

class DS18S20(OneWireTemp):
    def __init__(self, slave=None):
        OneWireTemp.__init__(self, slave, 0x10, "DS18S20")
        
class DS1822(OneWireTemp):
    def __init__(self, slave=None):
        OneWireTemp.__init__(self, slave, 0x22, "DS1822")
        
class DS18B20(OneWireTemp):
    def __init__(self, slave=None):
        OneWireTemp.__init__(self, slave, 0x28, "DS18B20")
        
class DS1825(OneWireTemp):
    def __init__(self, slave=None):
        OneWireTemp.__init__(self, slave, 0x3B, "DS1825")
        
class DS28EA00(OneWireTemp):
    def __init__(self, slave=None):
        OneWireTemp.__init__(self, slave, 0x42, "DS28EA00")





from rocket.daisy.utils.types import toint
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.analog import DAC, ADC


class PCF8591(DAC, I2C):
    def __init__(self, slave=0x48, vref=3.3):
        I2C.__init__(self, toint(slave))
        DAC.__init__(self, 5, 8, float(vref))
        self.daValue = 0
        
        
    def __str__(self):
        return "PCF8591(slave=0x%02X)" % self.slave
    
    def __family__(self):
        return [DAC.__family__(self), ADC.__family__(self)]

    def __command__(self, adChannel=0):
        d = bytearray(2)
        d[0]  = 0x40           # enable output
        d[0] |= adChannel & 0x03
        d[1]  = self.daValue
        return d
    

    def __analogRead__(self, channel, diff=False):
        if (channel == 0):
            return self.daValue
        else:
            self.writeBytes(self.__command__(channel-1))
            return self.readBytes(3)[2]
                
    
    def __analogWrite__(self, channel, value):
        if (channel == 0):
            self.daValue = value
            self.writeBytes(self.__command__())



# +---------------------------------------------------------------------------+
#
#      Program:    direction_converter.py
#
#      Purpose:    translation table
#                  referenced by the RC skidloader webiopi adapter				
#      
#      Target:     ARMV61A
#
#      Author:     Martin Shishkov
#
#      License:    GPL 3
# +---------------------------------------------------------------------------+

from enum import IntEnum
class directions(IntEnum):
    Ahead     = 0
    Left      = 1
    Right     = 2
    Stop      = 3
    Backward  = 4
    ArmUp     = 5
    ArmDown   = 6
    TiltUp    = 7
    TiltDown  = 8
    Lights	  = 9
    Camera	  = 10
    Sound	  = 11
    Parked	  = 12
    LR	  	  = 13


def directionToInt(direction):
    switcher = {
        "Ahead"     : directions.Ahead,
        "Left"      : directions.Left,
        "Right"     : directions.Right,
        "Stop"      : directions.Stop,
        "Backward"  : directions.Backward,
        "ArmUp"  	: directions.ArmUp,
        "ArmDown"   : directions.ArmDown,
        "TiltUp" 	: directions.TiltUp,
        "TiltDown" 	: directions.TiltDown,
        "Lights" 	: directions.Lights,
        "Camera" 	: directions.Camera,
        "Sound" 	: directions.Sound,
        "Parked" 	: directions.Parked,
        "LR" 		: directions.LR
    }
    return switcher.get(direction)
    
def directionToArray(direction):
    switcher = {
        "Ahead"     : [0x0,0xA,0x0,0x0,0xA,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xD],
        "Left"      : [0x1,0xA,0x0,0x0,0x8,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xD],
        "Right"     : [0x0,0xA,0x0,0x1,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xD],
        "Stop"      : [0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xD],
        "Backward"  : [0x1,0xA,0x0,0x1,0xA,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xD],
        "ArmUp"  	: [0x0,0x0,0x0,0x0,0x0,0x0,0xB,0x6,0x6,0x0,0x0,0x0,0x0,0x0,0xD],
        "ArmDown"   : [0x0,0x0,0x0,0x0,0x0,0x0,0x4,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xD],
        "TiltUp" 	: [0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x5,0x6,0x6,0x0,0x0,0xD],
        "TiltDown" 	: [0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xA,0xC,0xC,0x0,0x0,0xD],
        "Lights" 	: [0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xC,0xC,0xF,0xF,0xD],
        "Camera" 	: [0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xF,0xF,0xD],
        "Sound" 	: [0x0,0x0,0x0,0x0,0x0,0x6,0x0,0x6,0x0,0x0,0xC,0x0,0xC,0xC,0xD],
        "Parked" 	: [0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x3,0x0,0x0,0x0,0x0,0xD],
        "DI" 		: [0x0,0x0,0x0,0x0,0x0,0x0,0x5,0x6,0x6,0xA,0xC,0xC,0x0,0x0,0x0],
        "GG" 		: [0x0,0x0,0x0,0x0,0x0,0x0,0x5,0x6,0x6,0xA,0xC,0xC,0x0,0x0,0x0],
        "ER" 		: [0x0,0x0,0x0,0x0,0xE,0x0,0xA,0x0,0x0,0xC,0x0,0x0,0x0,0x0,0xD]
    }

    #high, low = byte >> 4, byte & 0x0F
    print ("Command:----------------- ", direction, " ------------------------")
    sequence = ("".join(hex(byte & 0x0F) for byte in switcher.get(direction)))
    sequence = sequence.replace("0x", "")
    sequence = sequence.upper()
    print (sequence)
    print ("-----------------------------------------------------------")
    return (sequence)
    
def directionToJoystick(direction, L : int, R : int):
    L = int(L)
    R = int(R)
    
    switcher = {
        "LR"      	: [0x0,L,0x0,0x0,R,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xD],
    }

    #high, low = byte >> 4, byte & 0x0F
    print ("Command:----------------- ", direction, " ------------------------")
    sequence = ("".join(hex(byte & 0x0F) for byte in switcher.get(direction)))
    sequence = sequence.replace("0x", "")
    sequence = sequence.upper()
    print (sequence)
    print ("-----------------------------------------------------------")
    return (sequence)

def intToDirection(i):
    switcher = {
		directions.Ahead 		:	"Ahead",
		directions.Left 		:	"Left",
		directions.Right 		:	"Right",
		directions.Stop 		:	"Stop",
		directions.Backward 	:	"Backward",
		directions.ArmUp 		:	"ArmUp",
		directions.ArmDown 		:	"ArmDown",
		directions.TiltUp 		:	"TiltUp",
		directions.TiltDown 	:	"TiltDown",
		directions.Lights 		:	"Lights",
		directions.Camera 		:	"Camera",
		directions.Sound 		:	"Sound",
		directions.Parked 		:	"Parked",
		directions.LR 			:	"LR"
    }
    return switcher.get(i)


import imp
from rocket.daisy.utils import logger
from rocket.daisy.utils import types
from rocket.daisy.devices.instance import DEVICES

from rocket.daisy.devices import serial, digital, analog, sensor, shield

PACKAGES = [serial, digital, analog, sensor, shield]
def findDeviceClass(name):
    for package in PACKAGES:
        if hasattr(package, name):
            return getattr(package, name)
        if hasattr(package, "DRIVERS"):
            for driver in package.DRIVERS:
                if name in package.DRIVERS[driver]:
                    (fp, pathname, stuff) = imp.find_module(package.__name__.replace(".", "/") + "/" + driver)
                    module = imp.load_module(driver, fp, pathname, stuff)
                    return getattr(module, name)
    return None

def addDevice(name, device, args):
    devClass = findDeviceClass(device)
    if devClass == None:
        raise Exception("Device driver not found for %s" % device)
    if len(args) > 0:
        dev = devClass(**args)
    else:
        dev = devClass()
    addDeviceInstance(name, dev, args)

def addDeviceInstance(name, dev, args):
    funcs = {"GET": {}, "POST": {}}
    for att in dir(dev):
        func = getattr(dev, att)
        if callable(func) and hasattr(func, "routed"):
            if name == "GPIO":
                logger.debug("Mapping %s.%s to REST %s /GPIO/%s" % (dev, att, func.method, func.path))
            else:
                logger.debug("Mapping %s.%s to REST %s /devices/%s/%s" % (dev, att, func.method, name, func.path))
            funcs[func.method][func.path] = func
    
    DEVICES[name] = {'device': dev, 'functions': funcs}
    if name == "GPIO":
        logger.info("GPIO - Native mapped to REST API /GPIO")
    else:
        logger.info("%s - %s mapped to REST API /devices/%s" % (dev.__family__(), dev, name))
        
def closeDevices():
    devices = [k for k in DEVICES.keys()]
    for name in devices:
        device = DEVICES[name]["device"]
        logger.debug("Closing device %s - %s" %  (name, device))
        del DEVICES[name]
        device.close()

def getDevicesJSON(compact=False):
    devname = "name"
    devtype = "type"
    
    devices = []
    for devName in DEVICES:
        if devName == "GPIO":
            continue
        instance = DEVICES[devName]["device"]
        if hasattr(instance, "__family__"):
            family = instance.__family__()
            if isinstance(family, str):
                devices.append({devname: devName, devtype:family})
            else:
                for fam in family:
                    devices.append({devname: devName, devtype:fam})
                    
        else:
            devices.append({devname: devName, type:instance.__str__()})

    return types.jsonDumps(sorted(devices, key=lambda dev: dev[devname]))






import os
from rocket.daisy.devices.bus import Bus, loadModule

EXTRAS = {
    "TEMP": {"loaded": False, "module": "w1-therm"},
    "2408": {"loaded": False, "module": "w1_ds2408"}

}

def loadExtraModule(name):
    if EXTRAS[name]["loaded"] == False:
        loadModule(EXTRAS[name]["module"])
        EXTRAS[name]["loaded"] = True

class OneWire(Bus):
    def __init__(self, slave=None, family=0, extra=None):
        Bus.__init__(self, "ONEWIRE", "/sys/bus/w1/devices/w1_bus_master1/w1_master_slaves", os.O_RDONLY)
        if self.fd > 0:
            os.close(self.fd)
            self.fd = 0

        self.family = family
        if  slave != None:
            addr = slave.split("-")
            if len(addr) == 1:
                self.slave = "%02x-%s" % (family, slave)
            elif len(addr) == 2:
                prefix = int(addr[0], 16)
                if family > 0 and family != prefix:
                    raise Exception("1-Wire slave address %s does not match family %02x" % (slave, family))
                self.slave = slave
        else:
            devices = self.deviceList()
            if len(devices) == 0:
                raise Exception("No device match family %02x" % family)
            self.slave = devices[0]

        loadExtraModule(extra)
        
    def __str__(self):
        return "1-Wire(slave=%s)" % self.slave
    
    def deviceList(self):
        devices = []
        with open(self.device) as f:
            lines = f.read().split("\n")
            if self.family > 0:
                prefix = "%02x-" % self.family
                for line in lines:
                    if line.startswith(prefix):
                        devices.append(line)
            else:
                devices = lines
        return devices;
    
    def read(self):
        with open("/sys/bus/w1/devices/%s/w1_slave" % self.slave) as f:
            data = f.read()
        return data




from rocket.daisy.utils.version import PYTHON_MAJOR

if PYTHON_MAJOR >= 3:
    import configparser as parser
else:
    import ConfigParser as parser

class Config():
    
    def __init__(self, configfile=None):
        config = parser.ConfigParser()
        config.optionxform = str
        if configfile != None:
            config.read(configfile)
        self.config = config

    def get(self, section, key, default):
        if self.config.has_option(section, key):
            return self.config.get(section, key)
        return default

    def getboolean(self, section, key, default):
        if self.config.has_option(section, key):
            return self.config.getboolean(section, key)
        return default
    
    def getint(self, section, key, default):
        if self.config.has_option(section, key):
            return self.config.getint(section, key)
        return default
    
    def items(self, section):
        if self.config.has_section(section):
            return self.config.items(section)
        return {}



# +-------------------------------------------------------------------------------+
#
#      Program:    rocket-daisy
#
#      Purpose:    setup for remote open control - key enabling technology (Rocket)
#
#      Target:     ARMV61A
#
#      Author:     Martin Shishkov
#
#      License:    GPL 3
# +-------------------------------------------------------------------------------+
import os
import sys
from rocket.daisy.server import Server
from rocket.daisy.utils.loader import loadScript
from rocket.daisy.utils.logger import exception, setDebug, info, logToFile
from rocket.daisy.utils.version import VERSION_STRING
from rocket.daisy.utils.thread import runLoop, stop

def displayHelp():
    print("Daisy command-line usage")
    print("daisy [-h] [-c config] [-l log] [-s script] [-d] [port]")
    print("")
    print("Options:")
    print("  -h, --help           Display this help")
    print("  -c, --config file    Load config from file")
    print("  -l, --log file       Log to file")
    print("  -s, --script file    Load script from file")
    print("  -d, --debug          Enable DEBUG")
    print("")
    print("Arguments:")
    print("  port                 Port to bind the HTTP Server")
    exit()

def main(argv):
    port = 8000
    configfile = None
    scriptfile = None
    logfile = None
    
    i = 1
    while i < len(argv):
        if argv[i] in ["-c", "-C", "--config-file"]:
            configfile = argv[i+1]
            i+=1
        elif argv[i] in ["-l", "-L", "--log-file"]:
            logfile = argv[i+1]
            i+=1
        elif argv[i] in ["-s", "-S", "--script-file"]:
            scriptfile = argv[i+1]
            i+=1
        elif argv[i] in ["-h", "-H", "--help"]:
            displayHelp()
        elif argv[i] in ["-d", "--debug"]:
            setDebug()
        else:
            try:
                port = int(argv[i])
            except ValueError:
                displayHelp()
        i+=1
    
    if logfile:
        logToFile(logfile)

    info("Starting %s" % VERSION_STRING)
    server = Server(port=port, configfile=configfile, scriptfile=scriptfile)
    runLoop()
    server.stop()

if __name__ == "__main__":
    try:
        main(sys.argv)
    except Exception as e:
        exception(e)
        stop()




from rocket.daisy.decorators.rest import request, response
from rocket.daisy.utils.types import M_JSON

class GPIOPort():
    IN  = 0
    OUT = 1
    
    LOW  = False
    HIGH = True
    
    def __init__(self, channelCount):
        self.digitalChannelCount = channelCount
        
    def checkDigitalChannel(self, channel):
        if not 0 <= channel < self.digitalChannelCount:
            raise ValueError("Channel %d out of range [%d..%d]" % (channel, 0, self.digitalChannelCount-1))

    def checkDigitalValue(self, value):
        if not (value == 0 or value == 1):
            raise ValueError("Value %d not in {0, 1}")
    

    @request("GET", "count")
    @response("%d")
    def digitalCount(self):
        return self.digitalChannelCount

    def __family__(self):
        return "GPIOPort"
    
    def __getFunction__(self, channel):
        raise NotImplementedError
    
    def __setFunction__(self, channel, func):
        raise NotImplementedError
    
    def __digitalRead__(self, chanel):
        raise NotImplementedError
        
    def __portRead__(self):
        raise NotImplementedError
    
    def __digitalWrite__(self, chanel, value):
        raise NotImplementedError
        
    def __portWrite__(self, value):
        raise NotImplementedError
    
    def getFunction(self, channel):
        self.checkDigitalChannel(channel)
        return self.__getFunction__(channel)  
    
    @request("GET", "%(channel)d/function")
    def getFunctionString(self, channel):
        func = self.getFunction(channel)
        if func == self.IN:
            return "IN"
        elif func == self.OUT:
            return "OUT"
#        elif func == GPIO.PWM:
#            return "PWM"
        else:
            return "UNKNOWN"
        
    def setFunction(self, channel, value):
        self.checkDigitalChannel(channel)
        self.__setFunction__(channel, value)
        return self.getFunction(channel)

    @request("POST", "%(channel)d/function/%(value)s")
    def setFunctionString(self, channel, value):
        value = value.lower()
        if value == "in":
            self.setFunction(channel, self.IN)
        elif value == "out":
            self.setFunction(channel, self.OUT)
#        elif value == "pwm":
#            self.setFunction(channel, GPIO.PWM)
        else:
            raise ValueError("Bad Function")
        return self.getFunctionString(channel)  

    @request("GET", "%(channel)d/value")
    @response("%d")
    def digitalRead(self, channel):
        self.checkDigitalChannel(channel)
        return self.__digitalRead__(channel)

    @request("GET", "*")
    @response(contentType=M_JSON)
    def wildcard(self, compact=False):
        if compact:
            f = "f"
            v = "v"
        else:
            f = "function"
            v = "value"
            
        values = {}
        for i in range(self.digitalChannelCount):
            if compact:
                func = self.getFunction(i)
            else:
                func = self.getFunctionString(i)
            values[i] = {f: func, v: int(self.digitalRead(i))}
        return values

    @request("GET", "*/integer")
    @response("%d")
    def portRead(self):
        return self.__portRead__()
    
    @request("POST", "%(channel)d/value/%(value)d")
    @response("%d")
    def digitalWrite(self, channel, value):
        self.checkDigitalChannel(channel)
        self.checkDigitalValue(value)
        self.__digitalWrite__(channel, value)
        return self.digitalRead(channel)  

    @request("POST", "*/integer/%(value)d")
    @response("%d")
    def portWrite(self, value):
        self.__portWrite__(value)
        return self.portRead()

DRIVERS = {}    
DRIVERS["mcp23XXX"] = ["MCP23008", "MCP23009", "MCP23017", "MCP23018", "MCP23S08", "MCP23S09", "MCP23S17", "MCP23S18"]
DRIVERS["pcf8574" ] = ["PCF8574", "PCF8574A"]
DRIVERS["ds2408" ] = ["DS2408"]




from rocket.daisy.utils.types import toint
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.analog import DAC


class MCP4725(DAC, I2C):
    def __init__(self, slave=0x60, vref=3.3):
        I2C.__init__(self, toint(slave))
        DAC.__init__(self, 1, 12, float(vref))
        
    def __str__(self):
        return "MCP4725(slave=0x%02X)" % self.slave

    def __analogRead__(self, channel, diff=False):
        d = self.readBytes(3)
        value = (d[1] << 8 | d[2]) >> 4
        return value
        
    
    def __analogWrite__(self, channel, value):
        d = bytearray(2)
        d[0] = (value >> 8) & 0x0F
        d[1] = value & 0xFF
        self.writeBytes(d)




from rocket.daisy.utils.types import toint
from rocket.daisy.devices.spi import SPI
from rocket.daisy.devices.analog import ADC

class MCP3X0X(SPI, ADC):
    def __init__(self, chip, channelCount, resolution, vref, name):
        SPI.__init__(self, toint(chip), 0, 8, 10000)
        ADC.__init__(self, channelCount, resolution, float(vref))
        self.name = name
        self.MSB_MASK = 2**(resolution-8) - 1

    def __str__(self):
        return "%s(chip=%d)" % (self.name, self.chip)

    def __analogRead__(self, channel, diff):
        data = self.__command__(channel, diff)
        r = self.xfer(data)
        return ((r[1] & self.MSB_MASK) << 8) | r[2]
    
class MCP300X(MCP3X0X):
    def __init__(self, chip, channelCount, vref, name):
        MCP3X0X.__init__(self, chip, channelCount, 10, vref, name)

    def __command__(self, channel, diff):
        d = [0x00, 0x00, 0x00]
        d[0] |= 1
        d[1] |= (not diff) << 7
        d[1] |= ((channel >> 2) & 0x01) << 6
        d[1] |= ((channel >> 1) & 0x01) << 5
        d[1] |= ((channel >> 0) & 0x01) << 4
        return d
        
class MCP3002(MCP300X):
    def __init__(self, chip=0, vref=3.3):
        MCP300X.__init__(self, chip, 2, vref, "MCP3002")

    def __analogRead__(self, channel, diff):
        data = self.__command__(channel, diff)
        r = self.xfer(data)

        # Format of return is
        # 1 empty bit
        # 1 null bit
        # 10 ADC bits
        return (r[0]<<14 | r[1]<<6 | r[2]>>2) & ((2**self._analogResolution)-1)

    def __command__(self, channel, diff):
        d = [0x00, 0x00, 0x00]
        d[0] |= 1                                 #start bit
        d[1] |= (not diff) << 7                   #single/differntial input
        d[1] |= channel  << 6                     #channel select
        return d

class MCP3004(MCP300X):
    def __init__(self, chip=0, vref=3.3):
        MCP300X.__init__(self, chip, 4, vref, "MCP3004")
        
class MCP3008(MCP300X):
    def __init__(self, chip=0, vref=3.3):
        MCP300X.__init__(self, chip, 8, vref, "MCP3008")
        
class MCP320X(MCP3X0X):
    def __init__(self, chip, channelCount, vref, name):
        MCP3X0X.__init__(self, chip, channelCount, 12, vref, name)

    def __command__(self, channel, diff):
        d = [0x00, 0x00, 0x00]
        d[0] |= 1 << 2
        d[0] |= (not diff) << 1
        d[0] |= (channel >> 2) & 0x01
        d[1] |= ((channel >> 1) & 0x01) << 7
        d[1] |= ((channel >> 0) & 0x01) << 6
        return d
    
class MCP3204(MCP320X):
    def __init__(self, chip=0, vref=3.3):
        MCP320X.__init__(self, chip, 4, vref, "MCP3204")
        
class MCP3208(MCP320X):
    def __init__(self, chip=0, vref=3.3):
        MCP320X.__init__(self, chip, 8, vref, "MCP3208")
        




#   Ti Sitara/Maxwell implementation 2020 Martin Shishkov - gulliversoft.com

import os
import socket
import threading
import codecs
import mimetypes as mime
import logging

from rocket.daisy.utils.version import VERSION_STRING, PYTHON_MAJOR
from rocket.daisy.utils.logger import info, exception, debug
from rocket.daisy.utils.crypto import encrypt
from rocket.daisy.utils.types import str2bool

if PYTHON_MAJOR >= 3:
    import http.server as BaseHTTPServer
else:
    import BaseHTTPServer


DAISY_DOCROOT = "/usr/share/daisy/htdocs"

class HTTPServer(BaseHTTPServer.HTTPServer, threading.Thread):
    if socket.has_ipv6:
        address_family = socket.AF_INET6

    def __init__(self, host, port, handler, context, docroot, index, auth=None, realm=None):
        try:
            BaseHTTPServer.HTTPServer.__init__(self, ("", port), HTTPHandler)
        except:
            self.address_family = socket.AF_INET
            BaseHTTPServer.HTTPServer.__init__(self, ("", port), HTTPHandler)

        threading.Thread.__init__(self, name="HTTPThread")
        self.host = host
        self.port = port

        if context:
            self.context = context
            if not self.context.startswith("/"):
                self.context = "/" + self.context
            if not self.context.endswith("/"):
                self.context += "/"
        else:
            self.context = "/"

        self.docroot = docroot

        if index:
            self.index = index
        else:
            self.index = "index.html"
            
        self.handler = handler
        self.auth = auth
        if (realm == None):
            self.authenticateHeader = "Basic realm=daisy"
        else:
            self.authenticateHeader = "Basic realm=%s" % realm

        self.running = True
        self.start()
            
    def get_request(self):
        sock, addr = self.socket.accept()
        sock.settimeout(10.0)
        return (sock, addr)

    def run(self):
        info("HTTP Server binded on http://%s:%s%s" % (self.host, self.port, self.context))
        try:
            self.serve_forever()
        except Exception as e:
            if self.running == True:
                exception(e)
        info("HTTP Server stopped")

    def stop(self):
        self.running = False
        self.server_close()

class HTTPHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    logger = logging.getLogger("HTTP")

    def log_message(self, fmt, *args):
        pass
    
    def log_error(self, fmt, *args):
        pass
        
    def version_string(self):
        return VERSION_STRING
    
    def checkAuthentication(self):
        if self.server.auth == None or len(self.server.auth) == 0:
            return True
        
        authHeader = self.headers.get('Authorization')
        if authHeader == None:
            return False
        
        if not authHeader.startswith("Basic "):
            return False
        
        auth = authHeader.replace("Basic ", "")
        if PYTHON_MAJOR >= 3:
            auth_hash = encrypt(auth.encode())
        else:
            auth_hash = encrypt(auth)
            
        if auth_hash == self.server.auth:
            return True
        return False

    def requestAuthentication(self):
        self.send_response(401)
        self.send_header("WWW-Authenticate", self.server.authenticateHeader)
        self.end_headers();
        
    
    def logRequest(self, code):
        debug('### daisy gulliversoft "%s %s %s" - %s %s (Client: %s <%s>)' % (self.command, self.path, self.request_version, code, self.responses[code][0], self.client_address[0], self.headers["User-Agent"]))
    
    def sendResponse(self, code, body=None, contentType="text/plain"):
        if code >= 400:
            if body != None:
                self.send_error(code, body)
            else:
                self.send_error(code)
        else:
            self.send_response(code)
            self.send_header("Cache-Control", "no-cache")
            if body != None:
                self.send_header("Content-Type", contentType);
                self.send_header("Content-Length", len(body));
                self.end_headers();
                self.wfile.write(body)
        self.logRequest(code)

    def findFile(self, filepath):
        if os.path.exists(filepath):
            if os.path.isdir(filepath):
                filepath += "/" + self.server.index
                if os.path.exists(filepath):
                    return filepath
            else:
                return filepath
        return None
        
                
    def serveFile(self, relativePath):
        if self.server.docroot != None:
            path = self.findFile(self.server.docroot + "/" + relativePath)
            if path == None:
                path = self.findFile("./" + relativePath)

        else:
            path = self.findFile("./" + relativePath)                
            if path == None:
                path = self.findFile(DAISY_DOCROOT + "/" + relativePath)

        if path == None and (relativePath.startswith("daisy.") or relativePath.startswith("jquery")):
            path = self.findFile(DAISY_DOCROOT + "/" + relativePath)

        if path == None:
            return self.sendResponse(404, "Not Found")

        realPath = os.path.realpath(path)
        
        if realPath.endswith(".py"):
            return self.sendResponse(403, "Not Authorized")
        
        if not (realPath.startswith(os.getcwd()) 
                or (self.server.docroot and realPath.startswith(self.server.docroot))
                or realPath.startswith(DAISY_DOCROOT)):
            return self.sendResponse(403, "Not Authorized")
        
        (contentType, encoding) = mime.guess_type(path)
        f = codecs.open(path, encoding=encoding)
        data = f.read().encode('utf-8')
        f.close()
        self.send_response(200)
        self.send_header("Content-Type", contentType);
        self.send_header("Content-Length", len(data))
        self.end_headers()
        debug("### daisy gulliversoft: data: %s serveFile: %s" % (data, realPath))
        self.wfile.write(data)
        self.logRequest(200)
        
    def processRequest(self):
        self.request.settimeout(None)
        if not self.checkAuthentication():
            return self.requestAuthentication()
        
        request = self.path.replace(self.server.context, "/").split('?')
        relativePath = request[0]
        if relativePath[0] == "/":
            relativePath = relativePath[1:]
            
        if relativePath == "daisy" or relativePath == "daisy/":
            self.send_response(301)
            self.send_header("Location", "/")
            self.end_headers()
            return

        params = {}
        if len(request) > 1:
            for s in request[1].split('&'):
                if s.find('=') > 0:
                    (name, value) = s.split('=')
                    params[name] = value
                else:
                    params[s] = None
        
        compact = False
        if 'compact' in params:
            compact = str2bool(params['compact'])

        try:
            result = (None, None, None)
            if self.command == "GET":
                result = self.server.handler.do_GET(relativePath, compact)
            elif self.command == "POST":
                length = 0
                length_header = 'content-length'
                if length_header in self.headers:
                    length = int(self.headers[length_header])
                result = self.server.handler.do_POST(relativePath, self.rfile.read(length), compact)
            else:
                result = (405, None, None)
                
            (code, body, contentType) = result
            
            if code > 0:
                self.sendResponse(code, body.encode('utf-8'), contentType)
            else:
                if self.command == "GET":
                    debug("### daisy gulliversoft: GET on on http://%s:%s%s file:%s" % (self.server.host, self.server.port, self.server.context, relativePath))
                    self.serveFile(relativePath)
                else:
                    self.sendResponse(404)

        except ValueError as e:
            self.sendResponse(403, "%s" % e)
        except Exception as e:
            self.sendResponse(500)
            raise e
            
    def do_GET(self):
        self.processRequest()

    def do_POST(self):
        self.processRequest()





import fcntl
import array
import ctypes
import struct

from rocket.daisy.utils.version import PYTHON_MAJOR
from rocket.daisy.devices.bus import Bus

# from spi/spidev.h
_IOC_NRBITS   =  8
_IOC_TYPEBITS =  8
_IOC_SIZEBITS = 14
_IOC_DIRBITS  =  2

_IOC_NRSHIFT    = 0
_IOC_TYPESHIFT  = (_IOC_NRSHIFT+_IOC_NRBITS)
_IOC_SIZESHIFT  = (_IOC_TYPESHIFT+_IOC_TYPEBITS)
_IOC_DIRSHIFT   = (_IOC_SIZESHIFT+_IOC_SIZEBITS)

_IOC_NONE   = 0
_IOC_WRITE  = 1
_IOC_READ   = 2

def _IOC(direction,t,nr,size):
    return (((direction)  << _IOC_DIRSHIFT) |
            ((size) << _IOC_SIZESHIFT) |
            ((t) << _IOC_TYPESHIFT) |
            ((nr)   << _IOC_NRSHIFT))
def _IOR(t, number, size):
    return _IOC(_IOC_READ, t, number, size)
def _IOW(t, number, size):
    return _IOC(_IOC_WRITE, t, number, size)

SPI_CPHA        = 0x01
SPI_CPOL        = 0x02

SPI_MODE_0      = (0|0) 
SPI_MODE_1      = (0|SPI_CPHA)
SPI_MODE_2      = (SPI_CPOL|0)
SPI_MODE_3      = (SPI_CPOL|SPI_CPHA)

# does not work
# SPI_CS_HIGH     = 0x04
# SPI_LSB_FIRST   = 0x08
# SPI_3WIRE       = 0x10
# SPI_LOOP        = 0x20
# SPI_NO_CS       = 0x40
# SPI_READY       = 0x80

SPI_IOC_MAGIC   = ord('k')

def SPI_IOC_MESSAGE(count):
    return _IOW(SPI_IOC_MAGIC, 0, count)

# Read / Write of SPI mode (SPI_MODE_0..SPI_MODE_3)
SPI_IOC_RD_MODE             = _IOR(SPI_IOC_MAGIC, 1, 1)
SPI_IOC_WR_MODE             = _IOW(SPI_IOC_MAGIC, 1, 1)

# Read / Write SPI bit justification
# does not work
# SPI_IOC_RD_LSB_FIRST        = _IOR(SPI_IOC_MAGIC, 2, 1)
# SPI_IOC_WR_LSB_FIRST        = _IOW(SPI_IOC_MAGIC, 2, 1)

# Read / Write SPI device word length (1..N)
SPI_IOC_RD_BITS_PER_WORD    = _IOR(SPI_IOC_MAGIC, 3, 1)
SPI_IOC_WR_BITS_PER_WORD    = _IOW(SPI_IOC_MAGIC, 3, 1)

# Read / Write SPI device default max speed hz
SPI_IOC_RD_MAX_SPEED_HZ     = _IOR(SPI_IOC_MAGIC, 4, 4)
SPI_IOC_WR_MAX_SPEED_HZ     = _IOW(SPI_IOC_MAGIC, 4, 4)

class SPI(Bus):
    def __init__(self, chip=0, mode=0, bits=8, speed=0):
        Bus.__init__(self, "SPI", "/dev/spidev0.%d" % chip)
        self.chip = chip

        val8 = array.array('B', [0])
        val8[0] = mode
        if fcntl.ioctl(self.fd, SPI_IOC_WR_MODE, val8):
            raise Exception("Cannot write SPI Mode")
        if fcntl.ioctl(self.fd, SPI_IOC_RD_MODE, val8):
            raise Exception("Cannot read SPI Mode")
        self.mode = struct.unpack('B', val8)[0]
        assert(self.mode == mode)

        val8[0] = bits
        if fcntl.ioctl(self.fd, SPI_IOC_WR_BITS_PER_WORD, val8):
            raise Exception("Cannot write SPI Bits per word")
        if fcntl.ioctl(self.fd, SPI_IOC_RD_BITS_PER_WORD, val8):
            raise Exception("Cannot read SPI Bits per word")
        self.bits = struct.unpack('B', val8)[0]
        assert(self.bits == bits)

        val32 = array.array('I', [0])
        if speed > 0:
            val32[0] = speed
            if fcntl.ioctl(self.fd, SPI_IOC_WR_MAX_SPEED_HZ, val32):
                raise Exception("Cannot write SPI Max speed")
        if fcntl.ioctl(self.fd, SPI_IOC_RD_MAX_SPEED_HZ, val32):
            raise Exception("Cannot read SPI Max speed")
        self.speed = struct.unpack('I', val32)[0]
        assert((self.speed == speed) or (speed == 0))
    
    def __str__(self):
        return "SPI(chip=%d, mode=%d, speed=%dHz)" % (self.chip, self.mode, self.speed)
        
    def xfer(self, txbuff=None):
        length = len(txbuff)
        if PYTHON_MAJOR >= 3:
            _txbuff = bytes(txbuff)
            _txptr = ctypes.create_string_buffer(_txbuff)
        else:
            _txbuff = str(bytearray(txbuff))
            _txptr = ctypes.create_string_buffer(_txbuff)
        _rxptr = ctypes.create_string_buffer(length)
        
        data = struct.pack("QQLLHBBL",  #64 64 32 32 16 8 8 32 b = 32B
                    ctypes.addressof(_txptr),
                    ctypes.addressof(_rxptr),
                    length,
                    self.speed,
                    0, #delay
                    self.bits,
                    0, # cs_change,
                    0  # pad
                    )
        
        fcntl.ioctl(self.fd, SPI_IOC_MESSAGE(len(data)), data)
        _rxbuff = ctypes.string_at(_rxptr, length)
        return bytearray(_rxbuff)
        




from rocket.daisy.devices.onewire import OneWire
from rocket.daisy.devices.digital import GPIOPort

class DS2408(OneWire, GPIOPort):
    FUNCTIONS = [GPIOPort.IN for i in range(8)]

    def __init__(self, slave=None):
        OneWire.__init__(self, slave, 0x29, "2408")
        GPIOPort.__init__(self, 8)
        self.portWrite(0x00)
    
    def __str__(self):
        return "DS2408(slave=%s)" % self.slave
    
    def __getFunction__(self, channel):
        return self.FUNCTIONS[channel]
      
    def __setFunction__(self, channel, value):
        if not value in [self.IN, self.OUT]:
            raise ValueError("Requested function not supported")
        self.FUNCTIONS[channel] = value
        if value == self.IN:
            self.__output__(channel, 0)

    def __digitalRead__(self, channel):
        mask = 1 << channel
        d = self.readState()
        if d != None:
            return (d & mask) == mask

        
    def __digitalWrite__(self, channel, value):
        mask = 1 << channel
        b = self.readByte()
        if value:
            b |= mask
        else:
            b &= ~mask
        self.writeByte(b)
        
    def __portWrite__(self, value):
        self.writeByte(value)
        
    def __portRead__(self):
        return self.readByte()
        
    def readState(self):
        try:
            with open("/sys/bus/w1/devices/%s/state" % self.slave, "rb") as f:
                data = f.read(1)
            return ord(data)
        except IOError:
            return -1

    def readByte(self):
        try:
            with open("/sys/bus/w1/devices/%s/output" % self.slave, "rb") as f:
                data = f.read(1)
            return bytearray(data)[0]
        except IOError:
            return -1
      
    def writeByte(self, value):
        try:
            with open("/sys/bus/w1/devices/%s/output" % self.slave, "wb") as f:
                f.write(bytearray([value]))
        except IOError:
                pass
            
        



#     ____        _           
#    / __ \____ _(_)______  __
#   / / / / __ `/ / ___/ / / /
#  / /_/ / /_/ / (__  ) /_/ / 
# /_____/\__,_/_/____/\__, /  
# martin shishkov    /____/   
#
#   Copyright 2020 Martin Shishkov, gulliversoft.com
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from time import sleep
 
from rocket.daisy.utils.version import BOARD_REVISION, VERSION
from rocket.daisy.utils.logger import setInfo, setDebug, info, debug, warn, error, exception
from rocket.daisy.utils.thread import runLoop
from rocket.daisy.server import Server
from rocket.daisy.devices.instance import deviceInstance
from rocket.daisy.decorators.rest import macro

from rocket.daisy.devices import bus as _bus


try:
    import _daisy.GPIO as GPIO
except:
    pass


setInfo()
_bus.checkAllBus()



import base64
import hashlib
from rocket.daisy.utils.version import PYTHON_MAJOR

def encodeCredentials(login, password):
    abcd = "%s:%s" % (login, password)
    if PYTHON_MAJOR >= 3:
        b = base64.b64encode(abcd.encode())
    else:
        b = base64.b64encode(abcd)
    return b

def encrypt(value):
    return hashlib.sha256(value).hexdigest()

def encryptCredentials(login, password):
    return encrypt(encodeCredentials(login, password))



def request(method="GET", path="", data=None):
    def wrapper(func):
        func.routed = True
        func.method = method
        func.path = path
        func.data = data
        return func
    return wrapper

def response(fmt="%s", contentType="text/plain"):
    def wrapper(func):
        func.format = fmt
        func.contentType = contentType
        return func
    return wrapper

def macro(func):
    func.macro = True
    return func




from rocket.daisy.utils.types import M_JSON
from rocket.daisy.devices.digital.mcp23XXX import MCP23S17
from rocket.daisy.decorators.rest import request, response


class PiFaceDigital():
    def __init__(self, board=0):
        mcp = MCP23S17(0, 0x20+board)
        mcp.writeRegister(mcp.getAddress(mcp.IODIR, 0), 0x00) # Port A as output
        mcp.writeRegister(mcp.getAddress(mcp.IODIR, 8), 0xFF) # Port B as input
        mcp.writeRegister(mcp.getAddress(mcp.GPPU,  0), 0x00) # Port A PU OFF
        mcp.writeRegister(mcp.getAddress(mcp.GPPU,  8), 0xFF) # Port B PU ON
        self.mcp = mcp
        self.board = board
        
    def __str__(self):
        return "PiFaceDigital(%d)" % self.board 

    def __family__(self):
        return "PiFaceDigital"
    
    def checkChannel(self, channel):
        if not channel in range(8):
            raise ValueError("Channel %d invalid" % channel)
    
    @request("GET", "digital/input/%(channel)d")
    @response("%d")
    def digitalRead(self, channel):
        self.checkChannel(channel)
        return not self.mcp.digitalRead(channel+8)
    
    @request("POST", "digital/output/%(channel)d/%(value)d")
    @response("%d")
    def digitalWrite(self, channel, value):
        self.checkChannel(channel)
        return self.mcp.digitalWrite(channel, value)
    
    @request("GET", "digital/output/%(channel)d")
    @response("%d")
    def digitalReadOutput(self, channel):
        self.checkChannel(channel)
        return self.mcp.digitalRead(channel)
    
    @request("GET", "digital/*")
    @response(contentType=M_JSON)
    def readAll(self):
        inputs = {}
        outputs = {}
        for i in range(8):
            inputs[i] = self.digitalRead(i)
            outputs[i] = self.digitalReadOutput(i)
        return {"input": inputs, "output": outputs}





import time
from rocket.daisy.utils.types import toint
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.analog import PWM

class PCA9685(PWM, I2C):
    MODE1    = 0x00
    PWM_BASE = 0x06
    PRESCALE = 0xFE
    
    M1_SLEEP    = 1<<4
    M1_AI       = 1<<5
    M1_RESTART  = 1<<7
    
    def __init__(self, slave=0x40, frequency=50):
        I2C.__init__(self, toint(slave))
        PWM.__init__(self, 16, 12, toint(frequency))
        self.VREF = 0
        
        self.prescale = int(25000000.0/((2**12)*self.frequency))
        self.mode1 = self.M1_RESTART | self.M1_AI
        
        self.writeRegister(self.MODE1, self.M1_SLEEP)
        self.writeRegister(self.PRESCALE, self.prescale)
        time.sleep(0.01)

        self.writeRegister(self.MODE1, self.mode1)
        
    def __str__(self):
        return "PCA9685(slave=0x%02X)" % self.slave

    def getChannelAddress(self, channel):
        return int(channel * 4 + self.PWM_BASE) 

    def __pwmRead__(self, channel):
        addr = self.getChannelAddress(channel)
        d = self.readRegisters(addr, 4)
        start = d[1] << 8 | d[0]
        end   = d[3] << 8 | d[2]
        return end-start
    
    def __pwmWrite__(self, channel, value):
        addr = self.getChannelAddress(channel)
        d = bytearray(4)
        d[0] = 0
        d[1] = 0
        d[2] = (value & 0x0FF)
        d[3] = (value & 0xF00) >> 8
        self.writeRegisters(addr, d)





from rocket.daisy.utils.types import M_JSON
from rocket.daisy.utils.logger import debug
from rocket.daisy.devices.digital import GPIOPort
from rocket.daisy.decorators.rest import request, response
try:
    import _daisy.GPIO as GPIO
except:
    pass

EXPORT = []

class NativeGPIO(GPIOPort):
    def __init__(self):
        GPIOPort.__init__(self, 54)
        self.export = range(54)
        self.post_value = True
        self.post_function = True
        self.gpio_setup = []
        self.gpio_reset = []
        
    def __str__(self):
        return "GPIO"
    
    def addGPIO(self, lst, gpio, params):
        gpio = int(gpio)
        params = params.split(" ")
        func = params[0].lower()
        if func == "in":
            func = GPIO.IN
        elif func == "out":
            func = GPIO.OUT
        else:
            raise Exception("Unknown function")
        
        value = -1
        if len(params) > 1:
            value = int(params[1])
        lst.append({"gpio": gpio, "func": func, "value": value})
    
    def addGPIOSetup(self, gpio, params):
        self.addGPIO(self.gpio_setup, gpio, params)
        
    def addGPIOReset(self, gpio, params):
        self.addGPIO(self.gpio_reset, gpio, params)
        
    def addSetups(self, gpios):
        for (gpio, params) in gpios:
            self.addGPIOSetup(gpio, params)

    def addResets(self, gpios):
        for (gpio, params) in gpios:
            self.addGPIOReset(gpio, params)
    
    def setup(self):
        for g in self.gpio_setup:
            gpio = g["gpio"]
            debug("Setup GPIO %d" % gpio)
            GPIO.setFunction(gpio, g["func"])
            if g["value"] >= 0 and GPIO.getFunction(gpio) == GPIO.OUT:
                GPIO.digitalWrite(gpio, g["value"])
    
    def close(self):
        for g in self.gpio_reset:
            gpio = g["gpio"]
            debug("Reset GPIO %d" % gpio)
            GPIO.setFunction(gpio, g["func"])
            if g["value"] >= 0 and GPIO.getFunction(gpio) == GPIO.OUT:
                GPIO.digitalWrite(gpio, g["value"])
        
    def checkDigitalChannelExported(self, channel):
        if not channel in self.export:
            raise GPIO.InvalidChannelException("Channel %d is not allowed" % channel)
        
    def checkPostingFunctionAllowed(self):
        if not self.post_function:
            raise ValueError("POSTing function to native GPIO not allowed")
    
    def checkPostingValueAllowed(self):
        if not self.post_value:
            raise ValueError("POSTing value to native GPIO not allowed")
    
    def __digitalRead__(self, channel):
        self.checkDigitalChannelExported(channel)
        return GPIO.digitalRead(channel)
    
    def __digitalWrite__(self, channel, value):
        self.checkDigitalChannelExported(channel)
        self.checkPostingValueAllowed()
        GPIO.digitalWrite(channel, value)

    def __getFunction__(self, channel):
        self.checkDigitalChannelExported(channel)
        return GPIO.getFunction(channel)
    
    def __setFunction__(self, channel, value):
        self.checkDigitalChannelExported(channel)
        self.checkPostingFunctionAllowed()
        GPIO.setFunction(channel, value)
        
    def __portRead__(self):
        value = 0
        for i in self.export:
            value |= GPIO.digitalRead(i) << i
        return value 
            
    def __portWrite__(self, value):
        if len(self.export) < 54:
            for i in self.export:
                if GPIO.getFunction(i) == GPIO.OUT:
                    GPIO.digitalWrite(i, (value >> i) & 1)
        else:
            raise Exception("Please limit exported GPIO to write integers")
            
    @request("GET", "*")
    @response(contentType=M_JSON)
    def wildcard(self, compact=False):
        if compact:
            f = "f"
            v = "v"
        else:
            f = "function"
            v = "value"
            
        values = {}
        print(self.export)
        for i in self.export:
            if compact:
                func = GPIO.getFunction(i)
            else:
                func = GPIO.getFunctionString(i)
            values[i] = {f: func, v: int(GPIO.digitalRead(i))}
        return values

    
    @request("GET", "%(channel)d/pulse", "%s")
    def getPulse(self, channel):
        self.checkDigitalChannelExported(channel)
        self.checkDigitalChannel(channel)
        return GPIO.getPulse(channel)
    
    @request("POST", "%(channel)d/sequence/%(args)s")
    @response("%d")
    def outputSequence(self, channel, args):
        self.checkDigitalChannelExported(channel)
        self.checkPostingValueAllowed()
        self.checkDigitalChannel(channel)
        (period, sequence) = args.split(",")
        period = int(period)
        GPIO.outputSequence(channel, period, sequence)
        return int(sequence[-1])
        
    @request("POST", "%(channel)d/pulse/")
    def pulse(self, channel):
        self.checkDigitalChannelExported(channel)
        self.checkPostingValueAllowed()
        self.checkDigitalChannel(channel)
        GPIO.pulse(channel)
        return "OK"
        
    @request("POST", "%(channel)d/pulseRatio/%(value)f")
    def pulseRatio(self, channel, value):
        self.checkDigitalChannelExported(channel)
        self.checkPostingValueAllowed()
        self.checkDigitalChannel(channel)
        GPIO.pulseRatio(channel, value)
        return GPIO.getPulse(channel)
        
    @request("POST", "%(channel)d/pulseAngle/%(value)f")
    def pulseAngle(self, channel, value):
        self.checkDigitalChannelExported(channel)
        self.checkPostingValueAllowed()
        self.checkDigitalChannel(channel)
        GPIO.pulseAngle(channel, value)
        return GPIO.getPulse(channel)
    



import time
import signal
import threading
from rocket.daisy.utils import logger

RUNNING = False
TASKS = []

class Task(threading.Thread):
    def __init__(self, func, loop=False):
        threading.Thread.__init__(self)
        self.func = func
        self.loop = loop
        self.running = True
        self.start()
    
    def stop(self):
        self.running = False

    def run(self):
        if self.loop:
            while self.running == True:
                self.func()
        else:
            self.func()

def stop(signum=0, frame=None):
    global RUNNING
    if RUNNING:
        logger.info("Stopping...")
        RUNNING = False
        for task in TASKS:
            task.stop()
                

def runLoop(func=None, async_=False):
    global RUNNING
    RUNNING = True
    signal.signal(signal.SIGINT, stop)
    signal.signal(signal.SIGTERM, stop)

    if func != None:
        if async_:
            TASKS.append(Task(func, True))
        else:
            while RUNNING:
                func()
    else:
        while RUNNING:
            time.sleep(1)





from rocket.daisy.utils.types import toint
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.digital import GPIOPort

class PCF8574(I2C, GPIOPort):
    FUNCTIONS = [GPIOPort.IN for i in range(8)]
    
    def __init__(self, slave=0x20):
        slave = toint(slave)
        if slave in range(0x20, 0x28):
            self.name = "PCF8574"
        elif slave in range(0x38, 0x40):
            self.name = "PCF8574A"
        else:
            raise ValueError("Bad slave address for PCF8574(A) : 0x%02X not in range [0x20..0x27, 0x38..0x3F]" % slave)
        
        I2C.__init__(self, slave)
        GPIOPort.__init__(self, 8)
        self.portWrite(0xFF)
        self.portRead()
    
    def __str__(self):
        return "%s(slave=0x%02X)" % (self.name, self.slave)
        
    def __getFunction__(self, channel):
        return self.FUNCTIONS[channel]
    
    def __setFunction__(self, channel, value):
        if not value in [self.IN, self.OUT]:
            raise ValueError("Requested function not supported")
        self.FUNCTIONS[channel] = value
        
    def __digitalRead__(self, channel):
        mask = 1 << channel
        d = self.readByte()
        return (d & mask) == mask 

    def __portRead__(self):
        return self.readByte()
    
    def __digitalWrite__(self, channel, value):
        mask = 1 << channel
        b = self.readByte()
        if value:
            b |= mask
        else:
            b &= ~mask
        self.writeByte(b)

    def __portWrite__(self, value):
        self.writeByte(value)
        
class PCF8574A(PCF8574):
    def __init__(self, slave=0x38):
        PCF8574.__init__(self, slave)
        




from rocket.daisy.utils.types import toint
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.sensor import Luminosity

class TSL_LIGHT_X(I2C, Luminosity):
    VAL_COMMAND = 0x80
    REG_CONTROL = 0x00 | VAL_COMMAND
    REG_CONFIG  = 0x01 | VAL_COMMAND

    VAL_PWON    = 0x03
    VAL_PWOFF   = 0x00
    VAL_INVALID = -1

    def __init__(self, slave, time, name="TSL_LIGHT_X"):
        I2C.__init__(self, toint(slave))
        self.name = name  
        self.wake() # devices are powered down after power reset, wake them
        self.setTime(toint(time))

    def __str__(self):
        return "%s(slave=0x%02X)" % (self.name, self.slave)

    def wake(self):
        self.__wake__()

    def __wake__(self):
        self.writeRegister(self.REG_CONTROL, self.VAL_PWON)

    def sleep(self):
        self.__sleep__()
        
    def __sleep__(self):
        self.writeRegister(self.REG_CONTROL, self.VAL_PWOFF)
        
    def setTime(self, time):
        self.__setTime__(time)

    def getTime(self):
        return self.__getTime__()
            
class TSL2561X(TSL_LIGHT_X):
    VAL_TIME_402_MS   = 0x02
    VAL_TIME_101_MS   = 0x01
    VAL_TIME_14_MS    = 0x00
        
    REG_CHANNEL_0_LOW = 0x0C | TSL_LIGHT_X.VAL_COMMAND
    REG_CHANNEL_1_LOW = 0x0E | TSL_LIGHT_X.VAL_COMMAND
    
    MASK_GAIN         = 0x10
    MASK_TIME         = 0x03
  
    def __init__(self, slave, time, gain, name="TSL2561X"):
        TSL_LIGHT_X.__init__(self, slave, time, name)             
        self.setGain(toint(gain))

    def __getLux__(self):
        ch0_bytes = self.readRegisters(self.REG_CHANNEL_0_LOW, 2)
        ch1_bytes = self.readRegisters(self.REG_CHANNEL_1_LOW, 2)
        ch0_word = ch0_bytes[1] << 8 | ch0_bytes[0]
        ch1_word = ch1_bytes[1] << 8 | ch1_bytes[0]
        if ch0_word == 0 | ch1_word == 0:
            return self.VAL_INVALID # Driver security, avoid crash in lux calculation
        else:
            scaling = self.time_multiplier * self.gain_multiplier
            return self.__calculateLux__(scaling * ch0_word, scaling * ch1_word)

    def setGain(self, gain):
        if gain == 1:
            bit_gain = 0
            self.gain_multiplier = 16
        elif gain == 16:
            bit_gain = 1
            self.gain_multiplier = 1
        else:
            raise ValueError("Gain %d out of range [%d,%d]" % (gain, 1, 16))
        new_byte_gain = (bit_gain << 4) & self.MASK_GAIN
        
        current_byte_config = self.readRegister(self.REG_CONFIG)
        new_byte_config = (current_byte_config & ~self.MASK_GAIN) | new_byte_gain
        self.writeRegister(self.REG_CONFIG, new_byte_config)
    
    def getGain(self):
        current_byte_config = self.readRegister(self.REG_CONFIG)
        if (current_byte_config & self.MASK_GAIN):
            return 16
        else:
            return 1

    def __setTime__(self, time):
        if not time in [14, 101, 402]:
            raise ValueError("Time %d out of range [%d,%d,%d]" % (time, 14, 101, 402))
        if time == 402:
            bits_time = self.VAL_TIME_402_MS
            self.time_multiplier = 1
        elif time == 101:
            bits_time = self.VAL_TIME_101_MS
            self.time_multiplier = 322 / 81
        elif time == 14:
            bits_time = self.VAL_TIME_14_MS
            self.time_multiplier = 322 / 11
        new_byte_time = bits_time & self.MASK_TIME

        current_byte_config = self.readRegister(self.REG_CONFIG)
        new_byte_config = (current_byte_config & ~self.MASK_TIME) | new_byte_time
        self.writeRegister(self.REG_CONFIG, new_byte_config)
        
    def __getTime__(self):
        current_byte_config = self.readRegister(self.REG_CONFIG)
        bits_time = (current_byte_config & self.MASK_TIME)
        if bits_time == self.VAL_TIME_402_MS:
            t = 402
        elif bits_time == self.VAL_TIME_101_MS:
            t = 101
        elif bits_time == self.VAL_TIME_14_MS:
            t =  14
        else:
            t = TSL_LIGHT_X.VAL_INVALID # indicates undefined
        return t
          
class TSL2561CS(TSL2561X):
    # Package CS (Chipscale) chip version
    def __init__(self, slave=0x39, time=402,  gain=1):
        TSL2561X.__init__(self, slave, time, gain, "TSL2561CS")

    def __calculateLux__(self, channel0_value, channel1value):
        channelRatio = channel1value / channel0_value
        if 0 < channelRatio <= 0.52:
            lux = 0.0315 * channel0_value - 0.0593 * channel0_value *(channelRatio**1.4)
        elif 0.52 < channelRatio <= 0.65:
            lux = 0.0229 * channel0_value - 0.0291 * channel1value
        elif 0.65 < channelRatio <= 0.80:
            lux = 0.0157 * channel0_value - 0.0180 * channel1value
        elif 0.80 < channelRatio <= 1.30:
            lux = 0.00338 * channel0_value - 0.00260 * channel1value
        else: # if channelRatio > 1.30
            lux = 0
        return lux
            
class TSL2561T(TSL2561X):
    # Package T (TMB-6)  chip version
    def __init__(self, slave=0x39, time=402, gain=1):
        TSL2561X.__init__(self, slave, time, gain, "TSL2561T")
        
    def __calculateLux__(self, channel0_value, channel1_value):
        channel_ratio = channel1_value / channel0_value
        if 0 < channel_ratio <= 0.50:
            lux = 0.0304 * channel0_value - 0.062 * channel0_value * (channel_ratio**1.4)
        elif 0.50 < channel_ratio <= 0.61:
            lux = 0.0224 * channel0_value - 0.031 * channel1_value
        elif 0.61 < channel_ratio <= 0.80:
            lux = 0.0128 * channel0_value - 0.0153 * channel1_value
        elif 0.80 < channel_ratio <= 1.30:
            lux = 0.00146 * channel0_value - 0.00112 * channel1_value
        else: # if channel_ratio > 1.30
            lux = 0
        return lux

class TSL2561(TSL2561T):
    # Default version for unknown packages, uses T Package class lux calculation
    def __init__(self, slave=0x39, time=402, gain=1):
        TSL2561X.__init__(self, slave, time, gain, "TSL2561")
        
        
class TSL4531(TSL_LIGHT_X):
    # Default version for unknown subtypes, uses 0x29 as slave address
    VAL_TIME_400_MS = 0x00
    VAL_TIME_200_MS = 0x01
    VAL_TIME_100_MS = 0x02

    REG_DATA_LOW    = 0x04 | TSL_LIGHT_X.VAL_COMMAND
    
    MASK_TCNTRL     = 0x03

    def __init__(self, slave=0x29, time=400, name="TSL4531"):
        TSL_LIGHT_X.__init__(self, slave, time, name)
        
    def __setTime__(self, time):
        if not time in [100, 200, 400]:
            raise ValueError("Time %d out of range [%d,%d,%d]" % (time, 100, 200, 400))
        if time == 400:
            bits_time = self.VAL_TIME_400_MS
            self.time_multiplier = 1
        elif time == 200:
            bits_time = self.VAL_TIME_200_MS
            self.time_multiplier = 2
        elif time == 100:
            bits_time = self.VAL_TIME_100_MS
            self.time_multiplier = 4            
        new_byte_time = bits_time & self.MASK_TCNTRL

        current_byte_config = self.readRegister(self.REG_CONFIG)
        new_byte_config = (current_byte_config & ~self.MASK_TCNTRL) | new_byte_time
        self.writeRegister(self.REG_CONFIG, new_byte_config)

    def __getTime__(self):
        current_byte_config =  self.readRegister(self.REG_CONFIG)
        bits_time = (current_byte_config & self.MASK_TCNTRL)
        if bits_time == self.VAL_TIME_400_MS:
            t = 400
        elif bits_time == self.VAL_TIME_200_MS:
            t = 200
        elif bits_time == self.VAL_TIME_100_MS:
            t = 100
        else:
            t = TSL_LIGHT_X.VAL_INVALID # indicates undefined
        return t

    def __getLux__(self):
        data_bytes = self.readRegisters(self.REG_DATA_LOW, 2)
        return self.time_multiplier * (data_bytes[1] << 8 | data_bytes[0])

class TSL45311(TSL4531):
    def __init__(self, slave=0x39, time=400):
        TSL4531.__init__(self, slave, time, "TSL45311")

class TSL45313(TSL4531):
    def __init__(self, slave=0x39, time=400):
        TSL4531.__init__(self, slave, time, "TSL45313")

class TSL45315(TSL4531):
    def __init__(self, slave=0x29, time=400):
        TSL4531.__init__(self, slave, time, "TSL45315")

class TSL45317(TSL4531):
    def __init__(self, slave=0x29, time=400):
        TSL4531.__init__(self, slave, time, "TSL45317")




#     ____        _           
#    / __ \____ _(_)______  __
#   / / / / __ `/ / ___/ / / /
#  / /_/ / /_/ / (__  ) /_/ / 
# /_____/\__,_/_/____/\__, /  
# martin shishkov    /____/   
#
#   Copyright 2020 Martin Shishkov, gulliversoft.com
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from rocket.daisy.utils import types
from rocket.daisy.utils import logger
from rocket.daisy.utils.types import M_JSON, M_PLAIN
from rocket.daisy.utils.version import BOARD_REVISION, VERSION_STRING, MAPPING
from rocket.daisy.devices import manager
from rocket.daisy.devices import instance
from rocket.daisy.devices.bus import BUSLIST

MACROS = {}

class RESTHandler():
    def __init__(self):
        self.device_mapping = True
        self.export = []
        self.routes = {}
        self.macros = {}
        

    def addMacro(self, macro):
        self.macros[macro.__name__] = macro
        
    def addRoute(self, source, destination):
        if source[0] == "/":
            source = source[1:]
        if destination[0] == "/":
            destination = destination[1:]
        self.routes[source] = destination
        logger.info("Added Route /%s => /%s" % (source, destination))
        
    def findRoute(self, path):
        for source in self.routes:
            if path.startswith(source):
                route = path.replace(source, self.routes[source])
                logger.info("Routing /%s => /%s" % (path, route))
                return route
        return path
        
    def extract(self, fmtArray, pathArray, args):
        if len(fmtArray) != len(pathArray):
            return False
        if len(fmtArray) == 0:
            return True
        fmt = fmtArray[0]
        path = pathArray[0]
        if fmt == path:
            return self.extract(fmtArray[1:], pathArray[1:], args)
        if fmt.startswith("%"):
            
            fmt = fmt[1:]
            t = 's'
            if fmt[0] == '(':
                if fmt[-1] == ')':
                    name = fmt[1:-1]
                elif fmt[-2] == ')':                                   
                    name = fmt[1:-2]
                    t = fmt[-1]
                else:
                    raise Exception("Missing closing brace")
            else:
                name = fmt
            
            if t == 's':
                args[name] = path
            elif t == 'b':
                args[name] = types.str2bool(path)
            elif t == 'd':
                args[name] = types.toint(path)
            elif t == 'x':
                args[name] = int(path, 16)
            elif t == 'f':
                args[name] = float(path)
            else:
                raise Exception("Unknown format type : %s" % t)
            
            return self.extract(fmtArray[1:], pathArray[1:], args)
            
        return False

    def getDeviceRoute(self, method, path):
        pathArray = path.split("/")
        deviceName = pathArray[0]
        device = instance.DEVICES[deviceName]
        if device == None:
            return (None, deviceName + " Not Found")
        pathArray = pathArray[1:]
        funcs = device["functions"][method]
        functionName = "/".join(pathArray)
        if functionName in funcs:
            return (funcs[functionName], {})
        
        for fname in funcs:
            func = funcs[fname]
            funcPathArray = func.path.split("/")
            args = {}
            if self.extract(funcPathArray, pathArray, args):
                return (func, args) 
        
        return (None, functionName + " Not Found")
    
    def callDeviceFunction(self, method, path, data=None):
        (func, args) = self.getDeviceRoute(method, path)
        if func == None:
            return (404, args, M_PLAIN)

        if func.data != None:
            args[func.data] = data
        
        result = func(**args)
        response = None
        contentType = None
        if result != None:
            if hasattr(func, "contentType"):
                contentType = func.contentType
                if contentType == M_JSON:
                    response = types.jsonDumps(result)
                else:
                    response = func.format % result
            else:
                response = result
        
        return (200, response, contentType)
        
    def do_GET(self, relativePath, compact=False):
        relativePath = self.findRoute(relativePath)
        
        # JSON full state
        if relativePath == "*":
            return (200, self.getJSON(compact), M_JSON)

        # server version
        elif relativePath == "version":
            return (200, VERSION_STRING, M_PLAIN)

        # board revision
        elif relativePath == "revision":
            revision = "%s" % BOARD_REVISION
            return (200, revision, M_PLAIN)

        
        elif relativePath == "devices/*":
            return (200, manager.getDevicesJSON(compact), M_JSON)
        
        elif relativePath.startswith("devices/"):
            if not self.device_mapping:
                return (404, None, None)
            path = relativePath.replace("devices/", "")
            return self.callDeviceFunction("GET", path)

        else:
            return (0, None, None)

    def do_POST(self, relativePath, data, compact=False):
        relativePath = self.findRoute(relativePath)

                
        if relativePath.startswith("macros/"):
            paths = relativePath.split("/")
            mname = paths[1]
            if len(paths) > 2:
                value = paths[2]
            else:
                value = ""

            if mname in self.macros:
                macro = self.macros[mname]

                if ',' in value:
                    args = value.split(',')
                    result = macro(*args)
                elif len(value) > 0:
                    result = macro(value)
                else:
                    result = macro()
                     
                response = ""
                if result:
                    response = "%s" % result
                return (200, response, M_PLAIN)
                    
            else:
                return (404, mname + " Not Found", M_PLAIN)
                
        elif relativePath.startswith("devices/"):
            if not self.device_mapping:
                return (404, None, None)
            path = relativePath.replace("devices/", "")
            return self.callDeviceFunction("POST", path, data)

        else: # path unknowns
            return (0, None, None)

    def getJSON(self, compact=False):
        if compact:
            f = 'f'
            v = 'v'
        else:
            f = 'function'
            v = 'value'
        
        json = {}
        for (bus, value) in BUSLIST.items():
            json[bus] = int(value["enabled"])
        
        gpios = {}
        json['GPIO'] = gpios
        return types.jsonDumps(json)





from time import sleep
from rocket.daisy.utils.types import toint, signInteger
from rocket.daisy.devices.i2c import I2C
from rocket.daisy.devices.analog import ADC


class ADS1X1X(ADC, I2C):
    VALUE     = 0x00
    CONFIG    = 0x01
    LO_THRESH = 0x02
    HI_THRESH = 0x03
    
    CONFIG_STATUS_MASK  = 0x80
    CONFIG_CHANNEL_MASK = 0x70
    CONFIG_GAIN_MASK    = 0x0E
    CONFIG_MODE_MASK    = 0x01
    
    def __init__(self, slave, channelCount, resolution, name):
        I2C.__init__(self, toint(slave))
        ADC.__init__(self, channelCount, resolution, 4.096)
        self._analogMax = 2**(resolution-1)
        self.name = name
        
        config = self.readRegisters(self.CONFIG, 2)
        
        mode = 0 # continuous
        config[0] &= ~self.CONFIG_MODE_MASK
        config[0] |= mode
        
        gain = 0x1 # FS = +/- 4.096V
        config[0] &= ~self.CONFIG_GAIN_MASK
        config[0] |= gain << 1
        
        self.writeRegisters(self.CONFIG, config)
    
    def __str__(self):
        return "%s(slave=0x%02X)" % (self.name, self.slave)
        
    def __analogRead__(self, channel, diff=False):
        config = self.readRegisters(self.CONFIG, 2)
        config[0] &= ~self.CONFIG_CHANNEL_MASK
        if diff:
            config[0] |= channel << 4
        else:
            config[0] |= (channel + 4) << 4
        self.writeRegisters(self.CONFIG, config)
        sleep(0.001)
        d = self.readRegisters(self.VALUE, 2)
        value = (d[0] << 8 | d[1]) >> (16-self._analogResolution)
        return signInteger(value, self._analogResolution)


class ADS1014(ADS1X1X):
    def __init__(self, slave=0x48):
        ADS1X1X.__init__(self, slave, 1, 12, "ADS1014")

class ADS1015(ADS1X1X):
    def __init__(self, slave=0x48):
        ADS1X1X.__init__(self, slave, 4, 12, "ADS1015")

class ADS1114(ADS1X1X):
    def __init__(self, slave=0x48):
        ADS1X1X.__init__(self, slave, 1, 16, "ADS1114")

class ADS1115(ADS1X1X):
    def __init__(self, slave=0x48):
        ADS1X1X.__init__(self, slave, 4, 16, "ADS1115")

